import graphlib
import time
import unittest

from tests.utility import get_available_tcp_port, logging_test_name

from scaler import Client, SchedulerClusterCombo
from scaler.utility.graph.optimization import cull_graph
from scaler.utility.logging.scoped_logger import ScopedLogger
from scaler.utility.logging.utility import setup_logger


def inc(i):
    return i + 1


def add(a, b):
    return a + b


def minus(a, b):
    return a - b


class TestGraph(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)
        self.address = f"tcp://127.0.0.1:{get_available_tcp_port()}"
        self.cluster = SchedulerClusterCombo(address=self.address, n_workers=3, event_loop="builtin")

    def tearDown(self) -> None:
        self.cluster.shutdown()
        pass

    def test_graph(self):
        graph = {"a": 2, "b": 2, "c": (inc, "a"), "d": (add, "a", "b"), "e": (minus, "d", "c")}

        with Client(self.address) as client:
            with ScopedLogger("test normal graph"):
                result = client.get(graph, ["e"])
                self.assertEqual(result, {"e": 1})

            with self.assertRaises(graphlib.CycleError):
                client.get({"b": (inc, "c"), "c": (inc, "b")}, ["b", "c"])

    def test_graph_fail(self):
        def inc_error(i):
            assert isinstance(i, int)
            time.sleep(1)
            raise ValueError("Compute Error")

        def add_sleep(a, b):
            time.sleep(5)
            return a + b

        graph = {"a": 2, "b": 2, "c": (inc_error, "a"), "d": (add_sleep, "a", "b"), "e": (minus, "d", "c")}

        with Client(self.address) as client:
            with ScopedLogger("test graph node fail"), self.assertRaises(ValueError):
                client.get(graph, ["e"])

            time.sleep(5)

        with Client(self.address) as client:
            import math

            graph["c"] = (math.sqrt, "a")

            with ScopedLogger("test graph node should restore from failure"):
                futures = client.get(graph, ["e"], block=False)
                futures["e"].result(timeout=15.0)

    def test_graph_return_order(self):
        graph = {
            "a": 2,
            "b": 2,
            "c": (inc, "a"),  # c = a + 1 = 2 + 1 = 3
            "d": (add, "a", "b"),  # d = a + b = 2 + 2 = 4
            "e": (minus, "d", "c"),  # e = d - c = 4 - 3 = 1
        }

        with Client(address=self.address) as client:
            with self.assertRaises(KeyError):
                client.get(graph, keys=["e", "d", "c", "c"])

            with ScopedLogger("test graph return order"):
                results = client.get(graph, keys=["e", "d", "c"])
                self.assertEqual(results, {"e": 1, "d": 4, "c": 3})

    def test_get_data_key(self):
        def func(a):
            return a

        with Client(self.address) as client:
            result = client.get({"a": (func, "b"), "b": [1]}, keys=["b"])
            self.assertEqual(result["b"], [1])

    def test_reference_data_key(self):
        with Client(address=self.address) as client:
            obj_ref = client.send_object(5, name="foobar")

            graph = {
                "a": 2,
                "b": obj_ref,
                "c": (inc, "a"),  # c = a + 1 = 2 + 1 = 3
                "d": (add, "a", "b"),  # d = a + b = 2 + 5 = 7
                "e": (minus, "d", "c"),  # e = d - c = 7 - 3 = 4
            }

            with ScopedLogger("test reference data key"):
                results = client.get(graph, keys=["e", "d", "c"])
                self.assertEqual(results, {"e": 4, "d": 7, "c": 3})

    def test_none_value(self):
        def func(a, optional_param_can_be_none):
            return a

        with Client(address=self.address) as client:
            graph = {"None": None, "a": 1, "b": (func, "a", "None")}

            with ScopedLogger("test None value in graph"):
                results = client.get(graph, keys=["b"])
                self.assertEqual(results, {"b": 1})

    def test_cancel(self):
        def func(a):
            time.sleep(10)
            return a

        with Client(address=self.address) as client:
            graph = {"a": 1, "b": (func, "a")}
            futures = client.get(graph, keys=["b"], block=False)

            time.sleep(4)
            futures["b"].cancel()

    def test_client_quit(self):
        def func(a):
            time.sleep(10)
            return a

        with Client(address=self.address) as client:
            graph = {"a": 1, "b": (func, "a")}
            futures = client.get(graph, keys=["b"], block=False)

            time.sleep(4)

        self.assertTrue(all(f.cancelled() for f in futures.values()))

    def test_cull_graph(self):
        graph = {
            "a": (lambda *_: None,),
            "b": (lambda *_: None, "a"),
            "c": (lambda *_: None, "a"),
            "d": (lambda *_: None, "b"),
            "e": (lambda *_: None, "b", "c"),
            "f": (lambda *_: None, "c"),
        }

        def filter_keys(_graph, keys):
            return {key: value for key, value in _graph.items() if key in keys}

        self.assertEqual(cull_graph(graph, ["d"]), filter_keys(graph, ["a", "b", "d"]))
        self.assertEqual(cull_graph(graph, ["e"]), filter_keys(graph, ["a", "b", "c", "e"]))
        self.assertEqual(cull_graph(graph, ["f"]), filter_keys(graph, ["a", "c", "f"]))
        self.assertEqual(cull_graph(graph, ["d", "e", "f"]), graph)



import logging
import os
import socket
import threading
import uuid
from typing import Optional

import zmq

from scaler.io.utility import deserialize, serialize
from scaler.protocol.python.mixins import Message
from scaler.utility.zmq_config import ZMQConfig


class SyncConnector:
    def __init__(self, context: zmq.Context, socket_type: int, address: ZMQConfig, identity: Optional[bytes]):
        self._address = address

        self._context = context
        self._socket = self._context.socket(socket_type)

        self._identity: bytes = (
            f"{os.getpid()}|{socket.gethostname().split('.')[0]}|{uuid.uuid4()}".encode()
            if identity is None
            else identity
        )

        # set socket option
        self._socket.setsockopt(zmq.IDENTITY, self._identity)
        self._socket.setsockopt(zmq.SNDHWM, 0)
        self._socket.setsockopt(zmq.RCVHWM, 0)

        self._socket.connect(self._address.to_address())

        self._lock = threading.Lock()

    def close(self):
        self._socket.close()

    @property
    def address(self) -> ZMQConfig:
        return self._address

    @property
    def identity(self) -> bytes:
        return self._identity

    def send(self, message: Message):
        with self._lock:
            self._socket.send(serialize(message), copy=False)

    def receive(self) -> Optional[Message]:
        with self._lock:
            payload = self._socket.recv(copy=False)

        return self.__compose_message(payload.bytes)

    def __compose_message(self, payload: bytes) -> Optional[Message]:
        result: Optional[Message] = deserialize(payload)
        if result is None:
            logging.error(f"{self.__get_prefix()}: received unknown message: {payload!r}")
            return None

        return result

    def __get_prefix(self):
        return f"{self.__class__.__name__}[{self._identity.decode()}]:"



import asyncio
import unittest

from tests.utility import logging_test_name

from scaler.utility.logging.utility import setup_logger
from scaler.utility.queues.async_sorted_priority_queue import AsyncSortedPriorityQueue


class TestSortedPriorityQueue(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)

    def test_sorted_priority_queue(self):
        async def async_test():
            queue = AsyncSortedPriorityQueue()

            await queue.put([2, 3])
            await queue.put([3, 5])
            await queue.put([1, 1])
            await queue.put([3, 6])
            await queue.put([2, 4])
            await queue.put([-3, 0])  # supports negative priorities
            await queue.put([1, 2])

            queue.remove(4)
            self.assertEqual(queue.qsize(), 6)

            self.assertEqual(await queue.get(), [-3, 0])
            self.assertEqual(await queue.get(), [1, 1])
            self.assertEqual(await queue.get(), [1, 2])
            self.assertEqual(await queue.get(), [2, 3])
            self.assertEqual(await queue.get(), [3, 5])
            self.assertEqual(await queue.get(), [3, 6])
            self.assertEqual(queue.qsize(), 0)
            self.assertTrue(not queue)
            self.assertTrue(queue.empty())

        asyncio.run(async_test())



import array
import concurrent.futures
import threading
from typing import Dict

import cloudpickle

try:
    import soamapi
except ImportError:
    raise ImportError("IBM Spectrum Symphony API not found, please install it with 'pip install soamapi'.")


class SoamMessage(soamapi.Message):
    def __init__(self, payload: bytes = b""):
        self.__payload = payload

    def set_payload(self, payload: bytes):
        self.__payload = payload

    def get_payload(self) -> bytes:
        return self.__payload

    def on_serialize(self, stream):
        payload_array = array.array("b", self.get_payload())
        stream.write_byte_array(payload_array, 0, len(payload_array))

    def on_deserialize(self, stream):
        self.set_payload(stream.read_byte_array("b"))


class SessionCallback(soamapi.SessionCallback):
    def __init__(self):
        self._callback_lock = threading.Lock()
        self._task_id_to_future: Dict[str, concurrent.futures.Future] = {}

    def on_response(self, task_output_handle):
        with self._callback_lock:
            task_id = task_output_handle.get_id()

            future = self._task_id_to_future.pop(task_id)

            if task_output_handle.is_successful():
                output_message = SoamMessage()
                task_output_handle.populate_task_output(output_message)
                result = cloudpickle.loads(output_message.get_payload())
                future.set_result(result)
            else:
                future.set_exception(task_output_handle.get_exception().get_embedded_exception())

    def on_exception(self, exception):
        with self._callback_lock:
            for future in self._task_id_to_future.values():
                future.set_exception(exception)

            self._task_id_to_future.clear()

    def submit_task(self, task_id: str, future: concurrent.futures.Future):
        self._task_id_to_future[task_id] = future

    def get_callback_lock(self) -> threading.Lock:
        return self._callback_lock



import dataclasses
import struct
from typing import Optional

from scaler.protocol.python.message import TaskResult


@dataclasses.dataclass
class ProfileResult:
    duration_s: float = dataclasses.field(default=0.0)
    memory_peak: int = dataclasses.field(default=0)
    cpu_time_s: float = dataclasses.field(default=0.0)

    FORMAT = "!fQf"  # duration, memory peak, CPU time

    def serialize(self) -> bytes:
        return struct.pack(self.FORMAT, self.duration_s, self.memory_peak, self.cpu_time_s)

    @staticmethod
    def deserialize(data: bytes) -> "ProfileResult":
        return ProfileResult(*struct.unpack(ProfileResult.FORMAT, data))


def retrieve_profiling_result_from_task_result(task_result: TaskResult) -> Optional[ProfileResult]:
    if task_result.metadata == b"":
        return None

    try:
        return ProfileResult.deserialize(task_result.metadata)
    except struct.error:
        raise ValueError(f"unexpected metadata value (expected {ProfileResult.__name__}).")



import dataclasses
import enum
import logging
import logging.config
import os
import typing


class LogType(enum.Enum):
    Screen = enum.auto()
    File = enum.auto()


@dataclasses.dataclass
class LogPath:
    log_type: LogType
    path: str


class LoggingLevel(enum.Enum):
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    NOTSET = logging.NOTSET


def setup_logger(
    log_paths: typing.Tuple[str, ...] = ("/dev/stdout",),
    logging_config_file: typing.Optional[str] = None,
    logging_level: str = LoggingLevel.INFO.name,
):
    if not log_paths and not logging_config_file:
        return

    if logging_config_file is not None:
        print(f"use logging config file: {logging_config_file}")
        logging.config.fileConfig(logging_config_file, disable_existing_loggers=True)
        return

    resolved_log_paths = [LogPath(log_type=__detect_log_types(file_name), path=file_name) for file_name in log_paths]
    __logging_config(log_paths=resolved_log_paths, logging_level=logging_level)
    logging.info(f"logging to {log_paths}")


def __detect_log_types(file_name: str) -> LogType:
    if file_name in {"-", "/dev/stdout"}:
        return LogType.Screen

    return LogType.File


def __format(name) -> str:
    if not name:
        return ""

    return "%({name})s".format(name=name)


def __generate_log_config() -> typing.Dict:
    return {
        "version": 1,
        "disable_existing_loggers": False,  # this fixes the problem
        "formatters": {
            "standard": {
                "format": "[{levelname}]{asctime}: {message}".format(
                    levelname=__format("levelname"), asctime=__format("asctime"), message=__format("message")
                ),
                "datefmt": "%Y-%m-%d %H:%M:%S%z",
            },
            "verbose": {
                "format": "[{levelname}]{asctime}:{module}:{funcName}:{lineno}: {message}".format(
                    levelname=__format("levelname"),
                    asctime=__format("asctime"),
                    module=__format("module"),
                    funcName=__format("funcName"),
                    lineno=__format("lineno"),
                    message=__format("message"),
                ),
                "datefmt": "%Y-%m-%d %H:%M:%S%z",
            },
        },
        "handlers": {},
        "loggers": {"": {"handlers": [], "level": "DEBUG", "propagate": True}},
    }


def __logging_config(log_paths: typing.List[LogPath], logging_level: str = LoggingLevel.INFO.name):
    logging.addLevelName(logging.INFO, "INFO")
    logging.addLevelName(logging.WARNING, "WARN")
    logging.addLevelName(logging.ERROR, "EROR")
    logging.addLevelName(logging.DEBUG, "DEBG")
    logging.addLevelName(logging.CRITICAL, "CTIC")

    config = __generate_log_config()
    handlers = config["handlers"]
    root_loggers = config["loggers"][""]["handlers"]

    for log_path in log_paths:
        if log_path.log_type == LogType.Screen:
            handlers["console"] = __create_stdout_handler(logging_level)
            root_loggers.append("console")
            continue

        elif log_path.log_type == LogType.File:
            handlers[log_path.path] = __create_time_rotating_file_handler(logging_level, log_path.path)
            root_loggers.append(log_path.path)
            continue

        raise TypeError(f"Unsupported LogPath: {log_path}")

    logging.config.dictConfig(config)


def __create_stdout_handler(logging_level: str):
    return {
        "class": "logging.StreamHandler",
        "level": logging_level,
        "formatter": "standard",
        "stream": "ext://sys.stdout",
    }


def __create_time_rotating_file_handler(logging_level: str, file_path: str):
    return {
        "class": "logging.handlers.TimedRotatingFileHandler",
        "level": logging_level,
        "formatter": "verbose",
        "filename": os.path.expandvars(os.path.expanduser(file_path)),
        "when": "midnight",
    }


def __create_size_rotating_file_handler(log_path) -> typing.Dict:
    return {
        "class": "logging.handlers.RotatingFileHandler",
        "level": "INFO",
        "formatter": "verbose",
        "filename": os.path.expandvars(os.path.expanduser(log_path)),
        "maxBytes": 10485760,
        "backupCount": 20,
        "encoding": "utf8",
    }


def __parse_logging_level(value):
    return LoggingLevel(value).value



import heapq
import sys
from asyncio import Queue
from typing import Dict, List, Tuple, Union

PriorityType = Union[int, Tuple["PriorityType", ...]]


class AsyncPriorityQueue(Queue):
    """A subclass of Queue; retrieves entries in priority order (lowest first).

    Entries are typically list of the form: [priority, data].
    """

    def __len__(self):
        return len(self._queue)

    def _init(self, maxsize):
        self._queue: List[List] = []
        self._locator: Dict[bytes, List] = {}

    def _put(self, item):
        if not isinstance(item, list):
            item = list(item)

        heapq.heappush(self._queue, item)
        self._locator[item[1]] = item

    def _get(self):
        priority, data = heapq.heappop(self._queue)
        self._locator.pop(data)
        return priority, data

    def remove(self, data):
        # this operation is O(n), first change priority to -1 and pop from top of the heap, mark it as invalid
        # entry in the heap is not good idea as those invalid, entry will never get removed, so we used heapq internal
        # function _siftdown to maintain min heap invariant
        item = self._locator.pop(data)
        i = self._queue.index(item)  # O(n)
        item[0] = self.__to_lowest_priority(item[0])
        heapq._siftdown(self._queue, 0, i)  # type: ignore[attr-defined]
        assert heapq.heappop(self._queue) == item

    def decrease_priority(self, data):
        # this operation should be O(n), mark it as invalid entry in the heap is not good idea as those invalid
        # entry will never get removed, so we used heapq internal function _siftdown to maintain min heap invariant
        item = self._locator[data]
        i = self._queue.index(item)  # O(n)
        item[0] = self.__to_lower_priority(item[0])
        heapq._siftdown(self._queue, 0, i)  # type: ignore[attr-defined]

    def max_priority(self):
        item = heapq.heappop(self._queue)
        heapq.heappush(self._queue, item)
        priority = item[0]
        return priority

    @classmethod
    def __to_lowest_priority(cls, original_priority: PriorityType) -> PriorityType:
        if isinstance(original_priority, tuple):
            return tuple(cls.__to_lowest_priority(value) for value in original_priority)
        else:
            return -sys.maxsize - 1

    @classmethod
    def __to_lower_priority(cls, original_priority: PriorityType) -> PriorityType:
        if isinstance(original_priority, tuple):
            return tuple(cls.__to_lower_priority(value) for value in original_priority)
        else:
            return original_priority - 1



import capnp  # noqa

import scaler.protocol.capnp.common_capnp as _common  # noqa
import scaler.protocol.capnp.message_capnp as _message  # noqa
import scaler.protocol.capnp.status_capnp as _status  # noqa



import logging
import multiprocessing
import os
import signal
from typing import Optional, Tuple

import psutil

from scaler.io.config import DEFAULT_PROCESSOR_KILL_DELAY_SECONDS
from scaler.protocol.python.message import Task
from scaler.utility.zmq_config import ZMQConfig
from scaler.worker.agent.processor.processor import SUSPEND_SIGNAL, Processor


class ProcessorHolder:
    def __init__(
        self,
        event_loop: str,
        address: ZMQConfig,
        garbage_collect_interval_seconds: int,
        trim_memory_threshold_bytes: int,
        hard_suspend: bool,
        logging_paths: Tuple[str, ...],
        logging_level: str,
    ):
        self._processor_id: Optional[bytes] = None
        self._task: Optional[Task] = None
        self._suspended = False

        self._hard_suspend = hard_suspend
        if hard_suspend:
            self._resume_event = None
            self._resumed_event = None
        else:
            context = multiprocessing.get_context("spawn")
            self._resume_event = context.Event()
            self._resumed_event = context.Event()

        self._processor = Processor(
            event_loop=event_loop,
            address=address,
            resume_event=self._resume_event,
            resumed_event=self._resumed_event,
            garbage_collect_interval_seconds=garbage_collect_interval_seconds,
            trim_memory_threshold_bytes=trim_memory_threshold_bytes,
            logging_paths=logging_paths,
            logging_level=logging_level,
        )
        self._processor.start()
        self._process = psutil.Process(self._processor.pid)

    def pid(self) -> int:
        assert self._processor.pid is not None
        return self._processor.pid

    def process(self) -> psutil.Process:
        return self._process

    def processor_id(self) -> bytes:
        assert self._processor_id is not None
        return self._processor_id

    def initialized(self) -> bool:
        return self._processor_id is not None

    def initialize(self, processor_id: bytes):
        self._processor_id = processor_id

    def task(self) -> Optional[Task]:
        return self._task

    def set_task(self, task: Optional[Task]):
        self._task = task

    def suspended(self) -> bool:
        return self._suspended

    def suspend(self):
        assert self._processor is not None
        assert self._task is not None
        assert self._suspended is False
        assert self.initialized()

        if self._hard_suspend:
            self.__send_signal("SIGSTOP")
        else:
            # If we do not want to hardly suspend the processor's process (e.g. to keep network links alive), we request
            # the process to wait on a synchronization event. That will stop the main thread while allowing the helper
            # threads to continue running.
            #
            # See https://github.com/Citi/scaler/issues/14

            assert self._resume_event is not None
            assert self._resumed_event is not None
            self._resume_event.clear()
            self._resumed_event.clear()

            self.__send_signal(SUSPEND_SIGNAL)

        self._suspended = True

    def resume(self):
        assert self._task is not None
        assert self._suspended is True

        if self._hard_suspend:
            self.__send_signal("SIGCONT")
        else:
            assert self._resume_event is not None
            assert self._resumed_event is not None

            self._resume_event.set()

            # Waits until the processor resumes processing. This avoids any future call to `suspend()` while the
            # processor hasn't returned from the `_resumed_event.wait()` call yet (causes a re-entrant error on Linux).
            self._resumed_event.wait()

        self._suspended = False

    def kill(self):
        self.__send_signal("SIGTERM")
        self._processor.join(DEFAULT_PROCESSOR_KILL_DELAY_SECONDS)

        if self._processor.exitcode is None:
            # TODO: some processors fail to interrupt because of a blocking 0mq call. Ideally we should interrupt
            # these blocking calls instead of sending a SIGKILL signal.

            logging.warning(f"Processor[{self.pid()}] does not terminate in time, send SIGKILL.")
            self.__send_signal("SIGKILL")
            self._processor.join()

        self.set_task(None)

    def __send_signal(self, signal_name: str):
        signal_instance = getattr(signal, signal_name, None)
        if signal_instance is None:
            raise RuntimeError(f"unsupported platform, signal not availaible: {signal_name}.")

        os.kill(self.pid(), signal_instance)



import argparse

from scaler.ui.webui import start_webui


def get_args():
    parser = argparse.ArgumentParser(
        "web ui for scaler monitoring", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("address", type=str, help="scheduler ipc address to connect to")
    parser.add_argument("--host", type=str, default="0.0.0.0", help="host for webserver to connect to")
    parser.add_argument("--port", type=int, default=50001, help="port for webserver to connect to")
    return parser.parse_args()


def main():
    args = get_args()
    start_webui(args.address, args.host, args.port)



import time
from typing import Optional

import psutil

from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.message import Resource, WorkerHeartbeat, WorkerHeartbeatEcho
from scaler.utility.mixins import Looper
from scaler.worker.agent.mixins import HeartbeatManager, TimeoutManager
from scaler.worker.symphony.task_manager import SymphonyTaskManager


class SymphonyHeartbeatManager(Looper, HeartbeatManager):
    def __init__(self):
        self._agent_process = psutil.Process()

        self._connector_external: Optional[AsyncConnector] = None
        self._worker_task_manager: Optional[SymphonyTaskManager] = None
        self._timeout_manager: Optional[TimeoutManager] = None

        self._start_timestamp_ns = 0
        self._latency_us = 0

    def register(
        self,
        connector_external: AsyncConnector,
        worker_task_manager: SymphonyTaskManager,
        timeout_manager: TimeoutManager,
    ):
        self._connector_external = connector_external
        self._worker_task_manager = worker_task_manager
        self._timeout_manager = timeout_manager

    async def on_heartbeat_echo(self, heartbeat: WorkerHeartbeatEcho):
        if self._start_timestamp_ns == 0:
            # not handling echo if we didn't send out heartbeat
            return

        self._latency_us = int(((time.time_ns() - self._start_timestamp_ns) / 2) // 1_000)
        self._start_timestamp_ns = 0
        self._timeout_manager.update_last_seen_time()

    async def routine(self):
        if self._start_timestamp_ns != 0:
            return

        await self._connector_external.send(
            WorkerHeartbeat.new_msg(
                Resource.new_msg(int(self._agent_process.cpu_percent() * 10), self._agent_process.memory_info().rss),
                psutil.virtual_memory().available,
                self._worker_task_manager.get_queued_size(),
                self._latency_us,
                self._worker_task_manager.can_accept_task(),
                [],
            )
        )
        self._start_timestamp_ns = time.time_ns()



import argparse
import asyncio
import functools
import signal

from scaler.io.config import (
    DEFAULT_CLIENT_TIMEOUT_SECONDS,
    DEFAULT_IO_THREADS,
    DEFAULT_LOAD_BALANCE_SECONDS,
    DEFAULT_LOAD_BALANCE_TRIGGER_TIMES,
    DEFAULT_MAX_NUMBER_OF_TASKS_WAITING,
    DEFAULT_OBJECT_RETENTION_SECONDS,
    DEFAULT_PER_WORKER_QUEUE_SIZE,
    DEFAULT_WORKER_TIMEOUT_SECONDS,
)
from scaler.scheduler.config import SchedulerConfig
from scaler.scheduler.scheduler import scheduler_main
from scaler.utility.event_loop import EventLoopType, register_event_loop
from scaler.utility.logging.utility import setup_logger
from scaler.utility.zmq_config import ZMQConfig


def get_args():
    parser = argparse.ArgumentParser("scaler scheduler", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--io-threads", type=int, default=DEFAULT_IO_THREADS, help="number of io threads for zmq")
    parser.add_argument(
        "--max-number-of-tasks-waiting",
        "-mt",
        type=int,
        default=DEFAULT_MAX_NUMBER_OF_TASKS_WAITING,
        help="max number of tasks can wait in scheduler while all workers are full",
    )
    parser.add_argument(
        "--client-timeout-seconds",
        "-ct",
        type=int,
        default=DEFAULT_CLIENT_TIMEOUT_SECONDS,
        help="discard client when timeout seconds reached",
    )
    parser.add_argument(
        "--worker-timeout-seconds",
        "-wt",
        type=int,
        default=DEFAULT_WORKER_TIMEOUT_SECONDS,
        help="discard worker when timeout seconds reached",
    )
    parser.add_argument(
        "--object-retention-seconds",
        "-ot",
        type=int,
        default=DEFAULT_OBJECT_RETENTION_SECONDS,
        help="discard function in scheduler when timeout seconds reached",
    )
    parser.add_argument(
        "--load-balance-seconds",
        "-ls",
        type=int,
        default=DEFAULT_LOAD_BALANCE_SECONDS,
        help="number of seconds for load balance operation in scheduler",
    )
    parser.add_argument(
        "--load-balance-trigger-times",
        "-lbt",
        type=int,
        default=DEFAULT_LOAD_BALANCE_TRIGGER_TIMES,
        help="exact number of repeated load balance advices when trigger load balance operation in scheduler",
    )
    parser.add_argument(
        "--per-worker-queue-size",
        "-qs",
        type=int,
        default=DEFAULT_PER_WORKER_QUEUE_SIZE,
        help="specify per worker queue size",
    )
    parser.add_argument(
        "--event-loop", "-e", default="builtin", choices=EventLoopType.allowed_types(), help="select event loop type"
    )
    parser.add_argument(
        "--protected", "-p", action="store_true", help="protect scheduler and worker from being shutdown by client"
    )
    parser.add_argument(
        "--logging-paths",
        "-lp",
        nargs="*",
        type=str,
        default=("/dev/stdout",),
        help="specify where scheduler log should logged to, it can accept multiple files, default is /dev/stdout",
    )
    parser.add_argument(
        "--logging-level",
        "-ll",
        type=str,
        choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
        default="INFO",
        help="specify the logging level",
    )
    parser.add_argument(
        "--logging-config-file",
        "-lc",
        type=str,
        default=None,
        help="use standard python the .conf file the specify python logging file configuration format, this will "
        "bypass --logging-path",
    )
    parser.add_argument("address", type=ZMQConfig.from_string, help="scheduler address to connect to")
    return parser.parse_args()


def main():
    args = get_args()
    setup_logger(args.logging_paths, args.logging_config_file, args.logging_level)

    scheduler_config = SchedulerConfig(
        event_loop=args.event_loop,
        address=args.address,
        io_threads=args.io_threads,
        max_number_of_tasks_waiting=args.max_number_of_tasks_waiting,
        per_worker_queue_size=args.per_worker_queue_size,
        client_timeout_seconds=args.client_timeout_seconds,
        worker_timeout_seconds=args.worker_timeout_seconds,
        object_retention_seconds=args.object_retention_seconds,
        load_balance_seconds=args.load_balance_seconds,
        load_balance_trigger_times=args.load_balance_trigger_times,
        protected=args.protected,
    )

    register_event_loop(args.event_loop)

    loop = asyncio.get_event_loop()
    __register_signal(loop)
    loop.run_until_complete(scheduler_main(scheduler_config))


def __register_signal(loop):
    loop.add_signal_handler(signal.SIGINT, functools.partial(__handle_signal))
    loop.add_signal_handler(signal.SIGTERM, functools.partial(__handle_signal))


def __handle_signal():
    for task in asyncio.all_tasks():
        task.cancel()



import math
import time
import unittest
from concurrent.futures import CancelledError, as_completed
from threading import Event

from tests.utility import get_available_tcp_port, logging_test_name

from scaler import Client, SchedulerClusterCombo
from scaler.utility.logging.utility import setup_logger


def noop_sleep(sec: int):
    time.sleep(sec)


class TestFuture(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)
        self.address = f"tcp://127.0.0.1:{get_available_tcp_port()}"
        self._workers = 3
        self.cluster = SchedulerClusterCombo(address=self.address, n_workers=self._workers, event_loop="builtin")

    def tearDown(self) -> None:
        self.cluster.shutdown()
        pass

    def test_callback(self):
        done_called_event = Event()

        def on_done_callback(fut):
            self.assertTrue(fut.done())
            self.assertAlmostEqual(fut.result(), 4.0)
            done_called_event.set()

        with Client(address=self.address) as client:
            fut = client.submit(math.sqrt, 16.0)
            fut.add_done_callback(on_done_callback)
            done_called_event.wait()  # wait for the callback to be called, DO NOT call result().

    def test_as_completed(self):
        with Client(address=self.address) as client:
            fut = client.submit(math.sqrt, 100.0)

            for finished in as_completed([fut], timeout=5):
                self.assertAlmostEqual(finished.result(), 10.0)

    def test_state(self):
        with Client(address=self.address) as client:
            fut = client.submit(noop_sleep, 0.5)
            self.assertTrue(fut.running())
            self.assertFalse(fut.done())

            fut.result()

            self.assertFalse(fut.running())
            self.assertTrue(fut.done())

    def test_cancel(self):
        with Client(address=self.address) as client:
            fut = client.submit(math.sqrt, 100.0)
            self.assertTrue(fut.cancel())

            self.assertTrue(fut.cancelled())
            self.assertTrue(fut.done())

            with self.assertRaises(CancelledError):
                fut.result()

            fut = client.submit(math.sqrt, 16)
            fut.result()

            # cancel() should fail on a completed future.
            self.assertFalse(fut.cancel())
            self.assertFalse(fut.cancelled())

    def test_exception(self):
        with Client(address=self.address) as client:
            fut = client.submit(math.sqrt, "16")

            with self.assertRaises(TypeError):
                fut.result()

            self.assertTrue(fut.done())

            self.assertIsInstance(fut.exception(), TypeError)

    def test_client_disconnected(self):
        with Client(address=self.address) as client:
            fut = client.submit(noop_sleep, 10.0)

        with self.assertRaises(CancelledError):
            fut.result()



from typing import Optional

import psutil

from scaler.io.async_binder import AsyncBinder
from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.message import StateScheduler
from scaler.protocol.python.status import Resource
from scaler.scheduler.mixins import ClientManager, ObjectManager, TaskManager, WorkerManager
from scaler.utility.mixins import Looper


class StatusReporter(Looper):
    def __init__(self, binder: AsyncConnector):
        self._monitor_binder: AsyncConnector = binder
        self._process = psutil.Process()

        self._binder: Optional[AsyncBinder] = None
        self._client_manager: Optional[ClientManager] = None
        self._object_manager: Optional[ObjectManager] = None
        self._task_manager: Optional[TaskManager] = None
        self._worker_manager: Optional[WorkerManager] = None

    def register_managers(
        self,
        binder: AsyncBinder,
        client_manager: ClientManager,
        object_manager: ObjectManager,
        task_manager: TaskManager,
        worker_manager: WorkerManager,
    ):
        self._binder = binder
        self._client_manager = client_manager
        self._object_manager = object_manager
        self._task_manager = task_manager
        self._worker_manager = worker_manager

    async def routine(self):
        await self._monitor_binder.send(
            StateScheduler.new_msg(
                binder=self._binder.get_status(),
                scheduler=Resource.new_msg(int(self._process.cpu_percent() * 10), self._process.memory_info().rss),
                rss_free=psutil.virtual_memory().available,
                client_manager=self._client_manager.get_status(),
                object_manager=self._object_manager.get_status(),
                task_manager=self._task_manager.get_status(),
                worker_manager=self._worker_manager.get_status(),
            )
        )



import asyncio
import logging
import uuid
from concurrent.futures import Future
from typing import Dict, Optional, Set, cast

import cloudpickle
from bidict import bidict

from scaler import Serializer
from scaler.io.async_connector import AsyncConnector
from scaler.io.utility import chunk_to_list_of_bytes, concat_list_of_bytes
from scaler.protocol.python.common import ObjectContent, TaskStatus
from scaler.protocol.python.message import (
    ObjectInstruction,
    ObjectRequest,
    ObjectResponse,
    Task,
    TaskCancel,
    TaskResult,
)
from scaler.utility.metadata.task_flags import retrieve_task_flags_from_task
from scaler.utility.mixins import Looper
from scaler.utility.object_utility import generate_object_id, generate_serializer_object_id, serialize_failure
from scaler.utility.queues.async_sorted_priority_queue import AsyncSortedPriorityQueue
from scaler.worker.agent.mixins import TaskManager
from scaler.worker.symphony.session_callback import SessionCallback, SoamMessage

try:
    import soamapi
except ImportError:
    raise ImportError("IBM Spectrum Symphony API not found, please install it with 'pip install soamapi'.")


class SymphonyTaskManager(Looper, TaskManager):
    def __init__(self, base_concurrency: int, service_name: str):
        if isinstance(base_concurrency, int) and base_concurrency <= 0:
            raise ValueError(f"base_concurrency must be a possible integer, got {base_concurrency}")

        self._base_concurrency = base_concurrency
        self._service_name = service_name

        self._executor_semaphore = asyncio.Semaphore(value=self._base_concurrency)

        self._task_id_to_task: Dict[bytes, Task] = dict()
        self._task_id_to_future: bidict[bytes, asyncio.Future] = bidict()

        self._process_task_lock = asyncio.Lock()  # to ensure the object cache is not accessed by multiple tasks
        self._serializers: Dict[bytes, Serializer] = dict()
        self._object_cache: Dict[bytes, bytes] = dict()  # this cache is ephemeral and is wiped after task processing

        # locks are used to wait for object responses to be received
        self._serializers_lock: Dict[bytes, asyncio.Lock] = dict()
        self._object_cache_lock: Dict[bytes, asyncio.Lock] = dict()

        self._queued_task_id_queue = AsyncSortedPriorityQueue()
        self._queued_task_ids: Set[bytes] = set()

        self._acquiring_task_ids: Set[bytes] = set()  # tasks contesting the semaphore
        self._processing_task_ids: Set[bytes] = set()
        self._canceled_task_ids: Set[bytes] = set()

        self._connector_external: Optional[AsyncConnector] = None

        """
        SOAM specific code
        """
        soamapi.initialize()

        self._session_callback = SessionCallback()

        self._ibm_soam_connection = soamapi.connect(
            self._service_name, soamapi.DefaultSecurityCallback("Guest", "Guest")
        )
        logging.info(f"established IBM Spectrum Symphony connection {self._ibm_soam_connection.get_id()}")

        ibm_soam_session_attr = soamapi.SessionCreationAttributes()
        ibm_soam_session_attr.set_session_type("RecoverableAllHistoricalData")
        ibm_soam_session_attr.set_session_name("ScalerSession")
        ibm_soam_session_attr.set_session_flags(soamapi.SessionFlags.PARTIAL_ASYNC)
        ibm_soam_session_attr.set_session_callback(self._session_callback)
        self._ibm_soam_session = self._ibm_soam_connection.create_session(ibm_soam_session_attr)
        logging.info(f"established IBM Spectrum Symphony session {self._ibm_soam_session.get_id()}")

    def register(self, connector: AsyncConnector):
        self._connector_external = connector

    async def routine(self):  # SymphonyTaskManager has two loops
        pass

    async def on_object_instruction(self, instruction: ObjectInstruction):
        pass

    async def on_object_response(self, response: ObjectResponse):
        if response.response_type != ObjectResponse.ObjectResponseType.Content:
            raise TypeError(f"invalid object response type received: {response.response_type}.")

        object_content = response.object_content

        for object_type, object_id, object_bytes in zip(
            object_content.object_types, object_content.object_ids, object_content.object_bytes
        ):
            if object_type == ObjectContent.ObjectContentType.Serializer:
                self._serializers[object_id] = cloudpickle.loads(concat_list_of_bytes(object_bytes))
                self._serializers_lock[object_id].release()
            elif object_type == ObjectContent.ObjectContentType.Object:
                self._object_cache[object_id] = concat_list_of_bytes(object_bytes)
                self._object_cache_lock[object_id].release()
            else:
                raise ValueError(f"invalid object type received: {object_type}")

    async def on_task_new(self, task: Task):
        task_priority = self.__get_task_priority(task)

        # if semaphore is locked, check if task is higher priority than all acquired tasks
        # if so, bypass acquiring and execute the task immediately
        if self._executor_semaphore.locked():
            for acquired_task_id in self._acquiring_task_ids:
                acquired_task = self._task_id_to_task[acquired_task_id]
                acquired_task_priority = self.__get_task_priority(acquired_task)
                if task_priority <= acquired_task_priority:
                    break
            else:
                self._task_id_to_task[task.task_id] = task
                self._processing_task_ids.add(task.task_id)
                self._task_id_to_future[task.task_id] = await self.__execute_task(task)
                return

        self._task_id_to_task[task.task_id] = task
        self._queued_task_id_queue.put_nowait((-task_priority, task.task_id))
        self._queued_task_ids.add(task.task_id)

    async def on_cancel_task(self, task_cancel: TaskCancel):
        task_queued = task_cancel.task_id in self._queued_task_ids
        task_processing = task_cancel.task_id in self._processing_task_ids

        if (not task_queued and not task_processing) or (task_processing and not task_cancel.flags.force):
            result = TaskResult.new_msg(task_cancel.task_id, TaskStatus.NotFound)
            await self._connector_external.send(result)
            return

        canceled_task = self._task_id_to_task[task_cancel.task_id]

        if task_queued:
            self._queued_task_ids.remove(task_cancel.task_id)
            self._queued_task_id_queue.remove(task_cancel.task_id)

            # task can be discarded because task was never submitted
            self._task_id_to_task.pop(task_cancel.task_id)

        if task_processing:
            future = self._task_id_to_future[task_cancel.task_id]
            future.cancel()

            # regardless of the future being canceled, the task is considered canceled and cleanup will occur later
            self._processing_task_ids.remove(task_cancel.task_id)
            self._canceled_task_ids.add(task_cancel.task_id)

        payload = [canceled_task.get_message().to_bytes()] if task_cancel.flags.retrieve_task_object else []
        result = TaskResult.new_msg(
            task_id=task_cancel.task_id, status=TaskStatus.Canceled, metadata=b"", results=payload
        )
        await self._connector_external.send(result)

    async def on_task_result(self, result: TaskResult):
        if result.task_id in self._queued_task_ids:
            self._queued_task_ids.remove(result.task_id)
            self._queued_task_id_queue.remove(result.task_id)

        self._processing_task_ids.remove(result.task_id)
        self._task_id_to_task.pop(result.task_id)

        await self._connector_external.send(result)

    def get_queued_size(self):
        return self._queued_task_id_queue.qsize()

    def can_accept_task(self):
        return not self._executor_semaphore.locked()

    async def resolve_tasks(self):
        if not self._task_id_to_future:
            return

        done, _ = await asyncio.wait(self._task_id_to_future.values(), return_when=asyncio.FIRST_COMPLETED)
        for future in done:
            task_id = self._task_id_to_future.inv.pop(future)
            task = self._task_id_to_task[task_id]

            if task_id in self._processing_task_ids:
                self._processing_task_ids.remove(task_id)

                if future.exception() is None:
                    serializer_id = generate_serializer_object_id(task.source)
                    serializer = self._serializers[serializer_id]

                    result_bytes = serializer.serialize(future.result())
                    status = TaskStatus.Success
                else:
                    result_bytes = serialize_failure(cast(Exception, future.exception()))
                    status = TaskStatus.Failed

                result_object_id = generate_object_id(task.source, uuid.uuid4().bytes)
                await self._connector_external.send(
                    ObjectInstruction.new_msg(
                        ObjectInstruction.ObjectInstructionType.Create,
                        task.source,
                        ObjectContent.new_msg(
                            object_ids=(result_object_id,),
                            object_types=(ObjectContent.ObjectContentType.Object,),
                            object_names=(f"<res {result_object_id.hex()[:6]}>".encode(),),
                            object_bytes=(chunk_to_list_of_bytes(result_bytes),),
                        ),
                    )
                )

                await self._connector_external.send(
                    TaskResult.new_msg(task_id, status, metadata=b"", results=[result_object_id])
                )

            elif task_id in self._canceled_task_ids:
                self._canceled_task_ids.remove(task_id)

            else:
                raise ValueError(f"task_id {task_id.hex()} not found in processing or canceled tasks")

            if task_id in self._acquiring_task_ids:
                self._acquiring_task_ids.remove(task_id)
                self._executor_semaphore.release()

            self._task_id_to_task.pop(task_id)

    async def process_task(self):
        await self._executor_semaphore.acquire()

        _, task_id = await self._queued_task_id_queue.get()
        task = self._task_id_to_task[task_id]

        self._acquiring_task_ids.add(task_id)
        self._processing_task_ids.add(task_id)
        self._task_id_to_future[task.task_id] = await self.__execute_task(task)

    async def __execute_task(self, task: Task) -> asyncio.Future:
        """
        This method is not very efficient because it does let objects linger in the cache. Each time inputs are
        requested, locks are acquired to wait for object responses to be received.
        """
        async with self._process_task_lock:

            awaited_locks = []
            request_ids = []

            serializer_id = generate_serializer_object_id(task.source)
            if serializer_id not in self._serializers:
                serializer_lock = asyncio.Lock()
                await serializer_lock.acquire()
                self._serializers_lock[serializer_id] = serializer_lock
                awaited_locks.append(serializer_lock)
                request_ids.append(serializer_id)

            object_ids = (task.func_object_id, *(arg.data for arg in task.function_args))
            for object_id in object_ids:
                object_lock = asyncio.Lock()
                await object_lock.acquire()
                self._object_cache_lock[object_id] = object_lock
                awaited_locks.append(object_lock)
                request_ids.append(object_id)

            await self._connector_external.send(
                ObjectRequest.new_msg(ObjectRequest.ObjectRequestType.Get, tuple(request_ids))
            )
            await asyncio.gather(*[awaited_lock.acquire() for awaited_lock in awaited_locks])

            serializer = self._serializers[serializer_id]

            function = serializer.deserialize(self._object_cache[task.func_object_id])
            args = [serializer.deserialize(self._object_cache[arg.data]) for arg in task.function_args]

            self._object_cache.clear()
            self._object_cache_lock.clear()
            self._serializers_lock.clear()

        """
        SOAM specific code
        """
        input_message = SoamMessage()
        input_message.set_payload(cloudpickle.dumps((function, *args)))

        task_attr = soamapi.TaskSubmissionAttributes()
        task_attr.set_task_input(input_message)

        with self._session_callback.get_callback_lock():
            symphony_task = self._ibm_soam_session.send_task_input(task_attr)

            future: Future = Future()
            future.set_running_or_notify_cancel()

            self._session_callback.submit_task(symphony_task.get_id(), future)

        return asyncio.wrap_future(future)

    @staticmethod
    def __get_task_priority(task: Task) -> int:
        priority = retrieve_task_flags_from_task(task).priority

        if priority < 0:
            raise ValueError(f"invalid task priority, must be positive or zero, got {priority}")

        return priority



import logging
import threading
from typing import Callable, Optional

import zmq

from scaler.io.utility import deserialize
from scaler.protocol.python.mixins import Message
from scaler.utility.zmq_config import ZMQConfig


class SyncSubscriber(threading.Thread):
    def __init__(
        self,
        address: ZMQConfig,
        callback: Callable[[Message], None],
        topic: bytes,
        exit_callback: Optional[Callable[[], None]] = None,
        stop_event: threading.Event = threading.Event(),
        daemonic: bool = False,
        timeout_seconds: int = -1,
    ):
        threading.Thread.__init__(self)

        self._stop_event = stop_event
        self._address = address
        self._callback = callback
        self._exit_callback = exit_callback
        self._topic = topic
        self.daemon = bool(daemonic)
        self._timeout_seconds = timeout_seconds

        self._context: Optional[zmq.Context] = None
        self._socket: Optional[zmq.Socket] = None

    def __close(self):
        self._socket.close()

    def __stop_polling(self):
        self._stop_event.set()

    def disconnect(self):
        self.__stop_polling()

    def run(self) -> None:
        self.__initialize()

        while not self._stop_event.is_set():
            self.__routine_polling()

        if self._exit_callback is not None:
            self._exit_callback()

        self.__close()

    def __initialize(self):
        self._context = zmq.Context.instance()
        self._socket = self._context.socket(zmq.SUB)
        self._socket.setsockopt(zmq.RCVHWM, 0)

        if self._timeout_seconds == -1:
            self._socket.setsockopt(zmq.RCVTIMEO, self._timeout_seconds)
        else:
            self._socket.setsockopt(zmq.RCVTIMEO, self._timeout_seconds * 1000)

        self._socket.subscribe(self._topic)
        self._socket.connect(self._address.to_address())
        self._socket.connect(self._address.to_address())

    def __routine_polling(self):
        try:
            self.__routine_receive(self._socket.recv(copy=False).bytes)
        except zmq.Again:
            raise TimeoutError(f"Cannot connect to {self._address.to_address()} in {self._timeout_seconds} seconds")

    def __routine_receive(self, payload: bytes):
        result: Optional[Message] = deserialize(payload)
        if result is None:
            logging.error(f"received unknown message: {payload!r}")
            return None

        self._callback(result)



import abc
from typing import Dict, List, Optional, Set


class TaskAllocator(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def add_worker(self, worker: bytes) -> bool:
        """add worker to worker collection"""
        raise NotImplementedError()

    @abc.abstractmethod
    def remove_worker(self, worker: bytes) -> List[bytes]:
        """remove worker to worker collection, and return list of task_ids of removed worker"""
        raise NotImplementedError()

    @abc.abstractmethod
    def get_worker_ids(self) -> Set[bytes]:
        """get all worker ids as list"""
        raise NotImplementedError()

    @abc.abstractmethod
    def get_worker_by_task_id(self, task_id: bytes) -> bytes:
        """get worker name by task id"""
        raise NotImplementedError()

    @abc.abstractmethod
    def balance(self) -> Dict[bytes, List[bytes]]:
        """balance worker, it should return list of task ids for over burdened worker, represented as worker
        identity to list of task ids dictionary"""
        raise NotImplementedError()

    @abc.abstractmethod
    async def assign_task(self, task_id: bytes) -> Optional[bytes]:
        """assign task_id in allocator, return None means no available worker, otherwise will return worker been
        assigned to"""
        raise NotImplementedError()

    @abc.abstractmethod
    def remove_task(self, task_id: bytes) -> Optional[bytes]:
        """remove task in allocator, return None means not found any worker, otherwise will return worker associate
        with the removed task_id"""
        raise NotImplementedError()

    @abc.abstractmethod
    def get_assigned_worker(self, task_id: bytes) -> Optional[bytes]:
        """get worker that been assigned to this task_id, return None means cannot find the worker assigned to this
        task id"""
        raise NotImplementedError()

    @abc.abstractmethod
    def has_available_worker(self) -> bool:
        """has available worker or not"""
        raise NotImplementedError()

    @abc.abstractmethod
    def statistics(self) -> Dict:
        raise NotImplementedError()



import asyncio
import functools
import logging

import zmq.asyncio

from scaler.io.async_binder import AsyncBinder
from scaler.io.async_connector import AsyncConnector
from scaler.io.config import CLEANUP_INTERVAL_SECONDS, STATUS_REPORT_INTERVAL_SECONDS
from scaler.protocol.python.message import (
    ClientDisconnect,
    ClientHeartbeat,
    DisconnectRequest,
    GraphTask,
    GraphTaskCancel,
    ObjectInstruction,
    ObjectRequest,
    Task,
    TaskCancel,
    TaskResult,
    WorkerHeartbeat,
)
from scaler.protocol.python.mixins import Message
from scaler.scheduler.client_manager import VanillaClientManager
from scaler.scheduler.config import SchedulerConfig
from scaler.scheduler.graph_manager import VanillaGraphTaskManager
from scaler.scheduler.object_manager import VanillaObjectManager
from scaler.scheduler.status_reporter import StatusReporter
from scaler.scheduler.task_manager import VanillaTaskManager
from scaler.scheduler.worker_manager import VanillaWorkerManager
from scaler.utility.event_loop import create_async_loop_routine
from scaler.utility.exceptions import ClientShutdownException
from scaler.utility.zmq_config import ZMQConfig, ZMQType


class Scheduler:
    def __init__(self, config: SchedulerConfig):
        if config.address.type != ZMQType.tcp:
            raise TypeError(
                f"{self.__class__.__name__}: scheduler address must be tcp type: {config.address.to_address()}"
            )

        self._address_monitor = ZMQConfig(
            type=ZMQType.ipc, host=f"/tmp/{config.address.host}_{config.address.port}_monitor"
        )

        logging.info(f"{self.__class__.__name__}: monitor address is {self._address_monitor.to_address()}")
        self._context = zmq.asyncio.Context(io_threads=config.io_threads)
        self._binder = AsyncBinder(context=self._context, name="scheduler", address=config.address)
        self._binder_monitor = AsyncConnector(
            context=self._context,
            name="scheduler_monitor",
            socket_type=zmq.PUB,
            address=self._address_monitor,
            bind_or_connect="bind",
            callback=None,
            identity=None,
        )
        self._client_manager = VanillaClientManager(
            client_timeout_seconds=config.client_timeout_seconds, protected=config.protected
        )
        self._object_manager = VanillaObjectManager()
        self._graph_manager = VanillaGraphTaskManager()
        self._task_manager = VanillaTaskManager(max_number_of_tasks_waiting=config.max_number_of_tasks_waiting)
        self._worker_manager = VanillaWorkerManager(
            per_worker_queue_size=config.per_worker_queue_size,
            timeout_seconds=config.worker_timeout_seconds,
            load_balance_seconds=config.load_balance_seconds,
            load_balance_trigger_times=config.load_balance_trigger_times,
        )
        self._status_reporter = StatusReporter(self._binder_monitor)

        # register
        self._binder.register(self.on_receive_message)
        self._client_manager.register(
            self._binder, self._binder_monitor, self._object_manager, self._task_manager, self._worker_manager
        )
        self._object_manager.register(self._binder, self._binder_monitor, self._client_manager, self._worker_manager)
        self._graph_manager.register(
            self._binder, self._binder_monitor, self._client_manager, self._task_manager, self._object_manager
        )
        self._task_manager.register(
            self._binder,
            self._binder_monitor,
            self._client_manager,
            self._object_manager,
            self._worker_manager,
            self._graph_manager,
        )
        self._worker_manager.register(self._binder, self._binder_monitor, self._task_manager)

        self._status_reporter.register_managers(
            self._binder, self._client_manager, self._object_manager, self._task_manager, self._worker_manager
        )

    async def on_receive_message(self, source: bytes, message: Message):
        # =====================================================================================
        # receive from upstream
        if isinstance(message, ClientHeartbeat):
            await self._client_manager.on_heartbeat(source, message)
            return

        if isinstance(message, GraphTask):
            await self._graph_manager.on_graph_task(source, message)
            return

        if isinstance(message, GraphTaskCancel):
            await self._graph_manager.on_graph_task_cancel(source, message)
            return

        if isinstance(message, Task):
            await self._task_manager.on_task_new(source, message)
            return

        if isinstance(message, TaskCancel):
            await self._task_manager.on_task_cancel(source, message)
            return

        # scheduler receives client shutdown request from upstream
        if isinstance(message, ClientDisconnect):
            await self._client_manager.on_client_disconnect(source, message)
            return

        # =====================================================================================
        # receive from downstream
        # receive worker heartbeat from downstream
        if isinstance(message, WorkerHeartbeat):
            await self._worker_manager.on_heartbeat(source, message)
            return

        # receive task result from downstream
        if isinstance(message, TaskResult):
            await self._worker_manager.on_task_result(message)
            return

        # scheduler receives worker disconnect request from downstream
        if isinstance(message, DisconnectRequest):
            await self._worker_manager.on_disconnect(source, message)
            return

        # =====================================================================================
        # object related
        # scheduler receives object request from upstream
        if isinstance(message, ObjectInstruction):
            await self._object_manager.on_object_instruction(source, message)
            return

        if isinstance(message, ObjectRequest):
            await self._object_manager.on_object_request(source, message)
            return

        logging.error(f"{self.__class__.__name__}: unknown message from {source=}: {message}")

    async def get_loops(self):
        loops = [
            create_async_loop_routine(self._binder.routine, 0),
            create_async_loop_routine(self._graph_manager.routine, 0),
            create_async_loop_routine(self._task_manager.routine, 0),
            create_async_loop_routine(self._client_manager.routine, CLEANUP_INTERVAL_SECONDS),
            create_async_loop_routine(self._object_manager.routine, CLEANUP_INTERVAL_SECONDS),
            create_async_loop_routine(self._worker_manager.routine, CLEANUP_INTERVAL_SECONDS),
            create_async_loop_routine(self._status_reporter.routine, STATUS_REPORT_INTERVAL_SECONDS),
        ]

        # if self._log_forwarder is not None:
        #     loops.append(create_async_loop_routine(self._log_forwarder.routine, 0))

        try:
            await asyncio.gather(*loops)
        except asyncio.CancelledError:
            pass
        except ClientShutdownException as e:
            logging.info(f"{self.__class__.__name__}: {e}")
            pass

        self._binder.destroy()
        self._binder_monitor.destroy()


@functools.wraps(Scheduler)
async def scheduler_main(*args, **kwargs):
    scheduler = Scheduler(*args, **kwargs)
    await scheduler.get_loops()



import time

from scaler.utility.mixins import Looper
from scaler.worker.agent.mixins import TimeoutManager


class VanillaTimeoutManager(Looper, TimeoutManager):
    def __init__(self, death_timeout_seconds: int):
        self._death_timeout_seconds = death_timeout_seconds
        self._last_seen_time = time.time()

    def update_last_seen_time(self):
        self._last_seen_time = time.time()

    async def routine(self):
        if (time.time() - self._last_seen_time) < self._death_timeout_seconds:
            return

        raise TimeoutError("timeout when connect to scheduler, quiting")



import logging
import time
import unittest

from tests.utility import get_available_tcp_port, logging_test_name

from scaler import Client, Cluster, SchedulerClusterCombo
from scaler.io.config import (
    DEFAULT_GARBAGE_COLLECT_INTERVAL_SECONDS,
    DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
    DEFAULT_IO_THREADS,
    DEFAULT_TASK_TIMEOUT_SECONDS,
    DEFAULT_TRIM_MEMORY_THRESHOLD_BYTES,
)
from scaler.utility.logging.utility import setup_logger
from scaler.utility.zmq_config import ZMQConfig

# This is a manual test because it can loop infinitely if it fails


class TestDeathTimeout(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)

    def test_no_scheduler(self):
        logging.info("test with no scheduler")
        # Test 1: Spinning up a cluster with no scheduler. Death timeout should apply
        cluster = Cluster(
            address=ZMQConfig.from_string(f"tcp://127.0.0.1:{get_available_tcp_port()}"),
            worker_io_threads=DEFAULT_IO_THREADS,
            worker_names=["a", "b"],
            heartbeat_interval_seconds=DEFAULT_HEARTBEAT_INTERVAL_SECONDS,
            garbage_collect_interval_seconds=DEFAULT_GARBAGE_COLLECT_INTERVAL_SECONDS,
            trim_memory_threshold_bytes=DEFAULT_TRIM_MEMORY_THRESHOLD_BYTES,
            task_timeout_seconds=DEFAULT_TASK_TIMEOUT_SECONDS,
            death_timeout_seconds=10,
            hard_processor_suspend=False,
            event_loop="builtin",
            logging_paths=("/dev/stdout",),
            logging_level="INFO",
            logging_config_file=None,
        )
        cluster.start()
        time.sleep(15)

    def test_shutdown(self):
        logging.info("test with explicitly shutdown")
        # Test 2: Running the Combo and sending shutdown
        address = f"tcp://127.0.0.1:{get_available_tcp_port()}"
        cluster = SchedulerClusterCombo(
            address=address, n_workers=2, per_worker_queue_size=2, event_loop="builtin", protected=False
        )
        client = Client(address=address)

        time.sleep(10)
        logging.info("Shutting down")
        client.shutdown()

        time.sleep(5)
        # this is combo cluster, client only shutdown clusters, not scheduler, so scheduler need be shutdown also
        cluster.shutdown()

    def test_no_timeout_if_suspended(self):
        """
        Client and scheduler shouldn't time out a client if it is running inside a suspended processor.
        """

        client_timeout_seconds = 3

        def parent(c: Client):
            return c.submit(child).result()

        def child():
            time.sleep(client_timeout_seconds + 1)  # prevents the parent task to execute.
            return "OK"

        address = f"tcp://127.0.0.1:{get_available_tcp_port()}"
        cluster = SchedulerClusterCombo(
            address=address,
            n_workers=1,
            per_worker_queue_size=2,
            event_loop="builtin",
            client_timeout_seconds=client_timeout_seconds,
        )

        try:
            with Client(address, timeout_seconds=client_timeout_seconds) as client:
                future = client.submit(parent, client)
                self.assertEqual(future.result(), "OK")
        finally:
            cluster.shutdown()



from typing import Dict, List

from scaler.protocol.capnp._python import _status  # noqa
from scaler.protocol.python.mixins import Message


class Resource(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def cpu(self) -> int:
        return self._msg.cpu

    @property
    def rss(self) -> int:
        return self._msg.rss

    @staticmethod
    def new_msg(cpu: int, rss: int) -> "Resource":  # type: ignore[override]
        return Resource(_status.Resource(cpu=cpu, rss=rss))

    def get_message(self):
        return self._msg


class ObjectManagerStatus(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def number_of_objects(self) -> int:
        return self._msg.numberOfObjects

    @property
    def object_memory(self) -> int:
        return self._msg.objectMemory

    @staticmethod
    def new_msg(number_of_objects: int, object_memory: int) -> "ObjectManagerStatus":  # type: ignore[override]
        return ObjectManagerStatus(
            _status.ObjectManagerStatus(numberOfObjects=number_of_objects, objectMemory=object_memory)
        )

    def get_message(self):
        return self._msg


class ClientManagerStatus(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def client_to_num_of_tasks(self) -> Dict[bytes, int]:
        return {p.client: p.numTask for p in self._msg.clientToNumOfTask}

    @staticmethod
    def new_msg(client_to_num_of_tasks: Dict[bytes, int]) -> "ClientManagerStatus":  # type: ignore[override]
        return ClientManagerStatus(
            _status.ClientManagerStatus(
                clientToNumOfTask=[
                    _status.ClientManagerStatus.Pair(client=p[0], numTask=p[1]) for p in client_to_num_of_tasks.items()
                ]
            )
        )

    def get_message(self):
        return self._msg


class TaskManagerStatus(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def unassigned(self) -> int:
        return self._msg.unassigned

    @property
    def running(self) -> int:
        return self._msg.running

    @property
    def success(self) -> int:
        return self._msg.success

    @property
    def failed(self) -> int:
        return self._msg.failed

    @property
    def canceled(self) -> int:
        return self._msg.canceled

    @property
    def not_found(self) -> int:
        return self._msg.notFound

    @staticmethod
    def new_msg(  # type: ignore[override]
        unassigned: int, running: int, success: int, failed: int, canceled: int, not_found: int
    ) -> "TaskManagerStatus":
        return TaskManagerStatus(
            _status.TaskManagerStatus(
                unassigned=unassigned,
                running=running,
                success=success,
                failed=failed,
                canceled=canceled,
                notFound=not_found,
            )
        )

    def get_message(self):
        return self._msg


class ProcessorStatus(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def pid(self) -> int:
        return self._msg.pid

    @property
    def initialized(self) -> int:
        return self._msg.initialized

    @property
    def has_task(self) -> bool:
        return self._msg.hasTask

    @property
    def suspended(self) -> bool:
        return self._msg.suspended

    @property
    def resource(self) -> Resource:
        return Resource(self._msg.resource)

    @staticmethod
    def new_msg(
        pid: int, initialized: int, has_task: bool, suspended: bool, resource: Resource  # type: ignore[override]
    ) -> "ProcessorStatus":
        return ProcessorStatus(
            _status.ProcessorStatus(
                pid=pid, initialized=initialized, hasTask=has_task, suspended=suspended, resource=resource.get_message()
            )
        )

    def get_message(self):
        return self._msg


class WorkerStatus(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def worker_id(self) -> bytes:
        return self._msg.workerId

    @property
    def agent(self) -> Resource:
        return Resource(self._msg.agent)

    @property
    def rss_free(self) -> int:
        return self._msg.rssFree

    @property
    def free(self) -> int:
        return self._msg.free

    @property
    def sent(self) -> int:
        return self._msg.sent

    @property
    def queued(self) -> int:
        return self._msg.queued

    @property
    def suspended(self) -> bool:
        return self._msg.suspended

    @property
    def lag_us(self) -> int:
        return self._msg.lagUS

    @property
    def last_s(self) -> int:
        return self._msg.lastS

    @property
    def itl(self) -> str:
        return self._msg.itl

    @property
    def processor_statuses(self) -> List[ProcessorStatus]:
        return [ProcessorStatus(ps) for ps in self._msg.processorStatuses]

    @staticmethod
    def new_msg(  # type: ignore[override]
        worker_id: bytes,
        agent: Resource,
        rss_free: int,
        free: int,
        sent: int,
        queued: int,
        suspended: int,
        lag_us: int,
        last_s: int,
        itl: str,
        processor_statuses: List[ProcessorStatus],
    ) -> "WorkerStatus":
        return WorkerStatus(
            _status.WorkerStatus(
                workerId=worker_id,
                agent=agent.get_message(),
                rssFree=rss_free,
                free=free,
                sent=sent,
                queued=queued,
                suspended=suspended,
                lagUS=lag_us,
                lastS=last_s,
                itl=itl,
                processorStatuses=[ps.get_message() for ps in processor_statuses],
            )
        )

    def get_message(self):
        return self._msg


class WorkerManagerStatus(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def workers(self) -> List[WorkerStatus]:
        return [WorkerStatus(ws) for ws in self._msg.workers]

    @staticmethod
    def new_msg(workers: List[WorkerStatus]) -> "WorkerManagerStatus":  # type: ignore[override]
        return WorkerManagerStatus(_status.WorkerManagerStatus(workers=[ws.get_message() for ws in workers]))

    def get_message(self):
        return self._msg


class BinderStatus(Message):
    def __init__(self, msg):
        super().__init__(msg)

    @property
    def received(self) -> Dict[str, int]:
        return {p.client: p.number for p in self._msg.received}

    @property
    def sent(self) -> Dict[str, int]:
        return {p.client: p.number for p in self._msg.sent}

    @staticmethod
    def new_msg(received: Dict[str, int], sent: Dict[str, int]) -> "BinderStatus":  # type: ignore[override]
        return BinderStatus(
            _status.BinderStatus(
                received=[_status.BinderStatus.Pair(client=p[0], number=p[1]) for p in received.items()],
                sent=[_status.BinderStatus.Pair(client=p[0], number=p[1]) for p in sent.items()],
            )
        )

    def get_message(self):
        return self._msg



from asyncio import Queue
from typing import Any, Dict

from scaler.utility.queues.async_priority_queue import AsyncPriorityQueue


class AsyncSortedPriorityQueue(Queue):
    """A subclass of Queue; retrieves entries in priority order (lowest first), and then by adding order.

    Entries are typically list of the form: [priority number, data].
    """

    def __len__(self):
        return len(self._queue)

    def _init(self, maxsize: int):
        self._queue = AsyncPriorityQueue()

        # Keeps an item count to assign monotonic integer to queued items, so to also keep the priority queue sorted by
        # adding order.
        # See https://docs.python.org/3/library/heapq.html#priority-queue-implementation-notes.
        self._item_counter: int = 0
        self._data_to_item_id: Dict[Any, int] = dict()

    def _put(self, item) -> None:
        priority, data = item

        if data in self._data_to_item_id:
            raise ValueError(f"item `{data}` already in the queue")

        item_id = self._item_counter
        self._item_counter += 1

        self._queue._put([priority, (item_id, data)])
        self._data_to_item_id[data] = item_id

    def _get(self):
        priority, (_, data) = self._queue._get()
        self._data_to_item_id.pop(data)

        return [priority, data]

    def remove(self, data: Any) -> None:
        item_id = self._data_to_item_id.pop(data)
        self._queue.remove((item_id, data))



import abc
from concurrent.futures import Future

from scaler.protocol.python.message import (
    ClientDisconnect,
    ClientHeartbeatEcho,
    ClientShutdownResponse,
    GraphTask,
    ObjectInstruction,
    ObjectRequest,
    ObjectResponse,
    Task,
    TaskCancel,
    TaskResult,
)


class HeartbeatManager(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def send_heartbeat(self):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_heartbeat_echo(self, heartbeat: ClientHeartbeatEcho):
        raise NotImplementedError()


class TimeoutManager(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def update_last_seen_time(self):
        raise NotImplementedError()


class ObjectManager(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def on_object_instruction(self, object_instruction: ObjectInstruction):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_object_request(self, request: ObjectRequest):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_task_result(self, result: TaskResult):
        raise NotImplementedError()

    @abc.abstractmethod
    async def clear_all_objects(self, clear_serializer: bool):
        raise NotImplementedError()


class TaskManager(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def on_new_task(self, task: Task):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_new_graph_task(self, task: GraphTask):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_task_result(self, result: TaskResult):
        raise NotImplementedError()


class FutureManager(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def add_future(self, future: Future):
        raise NotImplementedError()

    @abc.abstractmethod
    def cancel_all_futures(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def set_all_futures_with_exception(self, exception: Exception):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_task_result(self, result: TaskResult):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_cancel_task(self, task_cancel: TaskCancel):
        raise NotImplementedError()

    @abc.abstractmethod
    def on_object_response(self, response: ObjectResponse):
        raise NotImplementedError()


class DisconnectManager(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def on_client_disconnect(self, disconnect: ClientDisconnect):
        raise NotImplementedError()

    @abc.abstractmethod
    async def on_client_shutdown_response(self, response: ClientShutdownResponse):
        raise NotImplementedError()



import logging
from typing import List, Optional

from scaler.io.config import CAPNP_DATA_SIZE_LIMIT, CAPNP_MESSAGE_SIZE_LIMIT
from scaler.protocol.capnp._python import _message  # noqa
from scaler.protocol.python.message import PROTOCOL
from scaler.protocol.python.mixins import Message


def deserialize(data: bytes) -> Optional[Message]:
    with _message.Message.from_bytes(data, traversal_limit_in_words=CAPNP_MESSAGE_SIZE_LIMIT) as payload:
        if not hasattr(payload, payload.which()):
            logging.error(f"unknown message type: {payload.which()}")
            return None

        message = getattr(payload, payload.which())
        return PROTOCOL[payload.which()](message)


def serialize(message: Message) -> bytes:
    payload = _message.Message(**{PROTOCOL.inverse[type(message)]: message.get_message()})
    return payload.to_bytes()


def chunk_to_list_of_bytes(data: bytes) -> List[bytes]:
    # TODO: change to list of memoryview when capnp can support memoryview
    return [data[i : i + CAPNP_DATA_SIZE_LIMIT] for i in range(0, len(data), CAPNP_DATA_SIZE_LIMIT)]


def concat_list_of_bytes(data: List[bytes]) -> bytes:
    return bytearray().join(data)



from collections import deque
from typing import Any, Callable, Dict, List, Tuple, Union


def cull_graph(
    graph: Dict[str, Tuple[Union[Callable, Any], ...]], keys: List[str]
) -> Dict[str, Tuple[Union[Callable, Any], ...]]:
    queue = deque(keys)
    visited = set()
    for target_key in keys:
        visited.add(target_key)

    while queue:
        key = queue.popleft()

        task = graph[key]
        if not (isinstance(task, tuple) and task and callable(task[0])):
            continue

        dependencies = set(task[1:])
        for predecessor_key in dependencies:
            if predecessor_key in visited:
                continue
            visited.add(predecessor_key)
            queue.append(predecessor_key)

    return {key: graph[key] for key in visited}



import argparse
import curses
import functools
from typing import Dict, List, Literal, Union

from scaler.io.sync_subscriber import SyncSubscriber
from scaler.protocol.python.message import StateScheduler
from scaler.protocol.python.mixins import Message
from scaler.utility.formatter import (
    format_bytes,
    format_integer,
    format_microseconds,
    format_percentage,
    format_seconds,
)
from scaler.utility.zmq_config import ZMQConfig

SORT_BY_OPTIONS = {
    ord("n"): "worker",
    ord("C"): "agt_cpu",
    ord("M"): "agt_rss",
    ord("c"): "cpu",
    ord("m"): "rss",
    ord("F"): "rss_free",
    ord("f"): "free",
    ord("w"): "sent",
    ord("d"): "queued",
    ord("s"): "suspended",
    ord("l"): "lag",
}

SORT_BY_STATE: Dict[str, Union[str, bool]] = {"sort_by": "cpu", "sort_by_previous": "cpu", "sort_reverse": True}


def get_args():
    parser = argparse.ArgumentParser(
        "monitor scheduler as top like", formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("--timeout", "-t", type=int, default=5, help="timeout seconds")
    parser.add_argument("address", help="scheduler address to connect to")
    return parser.parse_args()


def main():
    args = get_args()
    curses.wrapper(poke, args)


def poke(screen, args):
    screen.nodelay(1)

    try:
        subscriber = SyncSubscriber(
            address=ZMQConfig.from_string(args.address),
            callback=functools.partial(show_status, screen=screen),
            topic=b"",
            daemonic=False,
            timeout_seconds=args.timeout,
        )
        subscriber.run()
    except KeyboardInterrupt:
        pass


def show_status(status: Message, screen):
    if not isinstance(status, StateScheduler):
        return

    __change_option_state(screen.getch())

    scheduler_table = __generate_keyword_data(
        "scheduler",
        {
            "cpu": format_percentage(status.scheduler.cpu),
            "rss": format_bytes(status.scheduler.rss),
            "rss_free": format_bytes(status.rss_free),
        },
    )

    task_manager_table = __generate_keyword_data(
        "task_manager",
        {
            "unassigned": status.task_manager.unassigned,
            "running": status.task_manager.running,
            "success": status.task_manager.success,
            "failed": status.task_manager.failed,
            "canceled": status.task_manager.canceled,
            "not_found": status.task_manager.not_found,
        },
        format_integer_flag=True,
    )
    object_manager = __generate_keyword_data(
        "object_manager",
        {
            "num_of_objs": status.object_manager.number_of_objects,
            "obj_mem": format_bytes(status.object_manager.object_memory),
        },
    )
    sent_table = __generate_keyword_data("scheduler_sent", status.binder.sent, format_integer_flag=True)
    received_table = __generate_keyword_data("scheduler_received", status.binder.received, format_integer_flag=True)
    client_table = __generate_keyword_data(
        "client_manager", status.client_manager.client_to_num_of_tasks, key_col_length=18
    )
    worker_manager_table = __generate_worker_manager_table(
        [
            {
                "worker": worker.worker_id.decode(),
                "agt_cpu": worker.agent.cpu,
                "agt_rss": worker.agent.rss,
                "cpu": sum(p.resource.cpu for p in worker.processor_statuses),
                "rss": sum(p.resource.rss for p in worker.processor_statuses),
                "os_rss_free": worker.rss_free,
                "free": worker.free,
                "sent": worker.sent,
                "queued": worker.queued,
                "suspended": worker.suspended,
                "lag": worker.lag_us,
                "last": worker.last_s,
                "ITL": worker.itl,
            }
            for worker in status.worker_manager.workers
        ],
        worker_length=24,
    )

    table1 = __merge_tables(scheduler_table, object_manager, padding="|")
    table1 = __merge_tables(table1, task_manager_table, padding="|")
    table1 = __merge_tables(table1, sent_table, padding="|")
    table1 = __merge_tables(table1, received_table, padding="|")

    table3 = __merge_tables(worker_manager_table, client_table, padding="|")

    screen.clear()
    try:
        new_row, max_cols = __print_table(screen, 0, table1, padding=1)
    except curses.error:
        __print_too_small(screen)
        return

    try:
        screen.addstr(new_row, 0, "-" * max_cols)
        screen.addstr(new_row + 1, 0, "Shortcuts: " + " ".join([f"{v}[{chr(k)}]" for k, v in SORT_BY_OPTIONS.items()]))
        screen.addstr(new_row + 3, 0, f"Total {len(status.worker_manager.workers)} worker(s)")
        _ = __print_table(screen, new_row + 4, table3)
    except curses.error:
        pass

    screen.refresh()


def __generate_keyword_data(title, data, key_col_length: int = 0, format_integer_flag: bool = False):
    table = [[title, ""]]

    def format_integer_func(value):
        if format_integer_flag:
            return format_integer(value)

        return value

    table.extend([[__truncate(k, key_col_length), format_integer_func(v)] for k, v in data.items()])
    return table


def __generate_worker_manager_table(wm_data: List[Dict], worker_length: int) -> List[List[str]]:
    if not wm_data:
        headers = [["No workers"]]
        return headers

    wm_data = sorted(
        wm_data, key=lambda item: item[SORT_BY_STATE["sort_by"]], reverse=bool(SORT_BY_STATE["sort_reverse"])
    )

    for row in wm_data:
        row["worker"] = __truncate(row["worker"], worker_length, how="left")
        row["agt_cpu"] = format_percentage(row["agt_cpu"])
        row["agt_rss"] = format_bytes(row["agt_rss"])
        row["cpu"] = format_percentage(row["cpu"])
        row["rss"] = format_bytes(row["rss"])
        row["os_rss_free"] = format_bytes(row["os_rss_free"])

        last = row.pop("last")
        last = f"({format_seconds(last)}) " if last > 5 else ""
        row["lag"] = last + format_microseconds(row["lag"])

    worker_manager_table = [[f"[{v}]" if v == SORT_BY_STATE["sort_by"] else v for v in wm_data[0].keys()]]
    worker_manager_table.extend([list(worker.values()) for worker in wm_data])
    return worker_manager_table


def __print_table(screen, line_number, data, padding: int = 1):
    if not data:
        return

    col_widths = [max(len(str(row[i])) for row in data) for i in range(len(data[0]))]

    for i, header in enumerate(data[0]):
        screen.addstr(line_number, sum(col_widths[:i]) + (padding * i), str(header).rjust(col_widths[i]))

    for i, row in enumerate(data[1:], start=1):
        for j, cell in enumerate(row):
            screen.addstr(line_number + i, sum(col_widths[:j]) + (padding * j), str(cell).rjust(col_widths[j]))

    return line_number + len(data), sum(col_widths) + (padding * len(col_widths))


def __merge_tables(left: List[List], right: List[List], padding: str = "") -> List[List]:
    if not left:
        return right

    if not right:
        return left

    result = []
    for i in range(max(len(left), len(right))):
        if i < len(left):
            left_row = left[i]
        else:
            left_row = [""] * len(left[0])

        if i < len(right):
            right_row = right[i]
        else:
            right_row = [""] * len(right[0])

        if padding:
            padding_column = [padding]
            result.append(left_row + padding_column + right_row)
        else:
            result.append(left_row + right_row)

    return result


def __concat_tables(up: List[List], down: List[List], padding: int = 1) -> List[List]:
    max_cols = max([len(row) for row in up] + [len(row) for row in down])
    for row in up:
        row.extend([""] * (max_cols - len(row)))

    padding_rows = [[""] * max_cols] * padding

    for row in down:
        row.extend([""] * (max_cols - len(row)))

    return up + padding_rows + down


def __truncate(string: str, number: int, how: Literal["left", "right"] = "left") -> str:
    if number <= 0:
        return string

    if len(string) <= number:
        return string

    if how == "left":
        return f"{string[:number]}+"
    else:
        return f"+{string[-number:]}"


def __print_too_small(screen):
    screen.clear()
    screen.addstr(0, 0, "Your terminal is too small to show")
    screen.refresh()


def __change_option_state(option: int):
    if option not in SORT_BY_OPTIONS.keys():
        return

    SORT_BY_STATE["sort_by_previous"] = SORT_BY_STATE["sort_by"]
    SORT_BY_STATE["sort_by"] = SORT_BY_OPTIONS[option]
    if SORT_BY_STATE["sort_by"] != SORT_BY_STATE["sort_by_previous"]:
        SORT_BY_STATE["sort_reverse"] = True
        return

    SORT_BY_STATE["sort_reverse"] = not SORT_BY_STATE["sort_reverse"]



import enum
import functools
import random
import unittest
from typing import Any

import cloudpickle
from tests.utility import get_available_tcp_port, logging_test_name

from scaler import Client, SchedulerClusterCombo, Serializer
from scaler.utility.logging.scoped_logger import ScopedLogger
from scaler.utility.logging.utility import setup_logger


def noop(sec: int):
    return sec * 1


def heavy_function(sec: int, payload: bytes):
    return len(payload) * sec


class ObjectType(enum.Enum):
    Integer = b"I"
    Other = b"O"


class MySerializer(Serializer):
    @staticmethod
    def serialize(obj: Any) -> bytes:
        print(f"MySerializer.serialize({MySerializer.trim_message_internal(obj)})")
        if isinstance(obj, int):
            return ObjectType.Integer.value + str(obj).encode()
        return ObjectType.Other.value + cloudpickle.dumps(obj)

    @staticmethod
    def deserialize(payload: bytes) -> Any:
        print(f"MySerializer.deserialize({MySerializer.trim_message_internal(payload)})")
        if payload[0] == ObjectType.Integer.value[0]:
            return int(payload[1:])
        assert payload[0] == ObjectType.Other.value[0]
        return cloudpickle.loads(payload[1:])

    @staticmethod
    def trim_message_internal(message: Any) -> str:
        msg = str(message)
        if len(msg) > 30:
            return f"{msg[:30]}..."
        return msg


class TestSerializer(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)
        self.address = f"tcp://127.0.0.1:{get_available_tcp_port()}"
        self._workers = 3
        self.cluster = SchedulerClusterCombo(address=self.address, n_workers=self._workers, event_loop="builtin")

    def tearDown(self) -> None:
        self.cluster.shutdown()
        pass

    def test_one_task(self):
        client = Client(self.address, serializer=MySerializer())

        with ScopedLogger("submitting task"):
            future = client.submit(noop, 1)

        with ScopedLogger("gathering task"):
            result = future.result()

        self.assertEqual(result, 1)
        print("done test_one_task")

    def test_heavy_function(self):
        with Client(self.address, serializer=MySerializer()) as client:
            size = 500_000_000
            tasks = [random.randint(0, 100) for _ in range(10000)]
            function = functools.partial(heavy_function, payload=b"1" * size)
            with ScopedLogger(f"submit {len(tasks)} heavy function (500mb) tasks"):
                results = client.map(function, [(i,) for i in tasks])

            expected = [task * size for task in tasks]
            self.assertEqual(results, expected)



import abc
from typing import Any


class Serializer(metaclass=abc.ABCMeta):
    @staticmethod
    @abc.abstractmethod
    def serialize(obj: Any) -> bytes:
        """
        Serialize the object to bytes, this serialization method is used to call for function object and EACH argument
        object and function result object, for example:

        def add(a, b):
            return a + b

        client.submit(add, 1, 2)

        The add function and the arguments 1 and 2 will be serialized and sent to the worker, and the result of the a+b
        will be serialized and sent back to the client, client will use deserialize function below to deserialize

        :param obj: the object to be serialized, can be function object, argument object, or function result object
        :return: serialized bytes of the object
        """

        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def deserialize(payload: bytes) -> Any:
        """
        Deserialize the bytes to the original object, this de-serialize method is used to deserialize the function
        object bytes and EACH serialized argument and serialized function result.

        :param payload: the serialized bytes of the object, can be function object, argument object, or function result
            object
        :return: any deserialized object
        """
        raise NotImplementedError()



import random

from dask.distributed import Client, LocalCluster

from scaler.utility.logging.scoped_logger import ScopedLogger
from scaler.utility.logging.utility import setup_logger


def sleep_print(sec: int):
    return sec * 1


def main():
    setup_logger()
    tasks = [random.randint(0, 100) for _ in range(10000)]

    cluster = LocalCluster(n_workers=10, threads_per_worker=2, memory_limit="100GB")
    client = Client(address=cluster.scheduler_address)

    with ScopedLogger(f"submit {len(tasks)} tasks"):
        futures = [client.submit(sleep_print, i) for i in tasks]

    with ScopedLogger(f"gather {len(futures)} results"):
        results = [future.result() for future in futures]

    assert results == tasks


if __name__ == "__main__":
    main()



from typing import Optional

from scaler.client.agent.mixins import DisconnectManager
from scaler.io.async_connector import AsyncConnector
from scaler.protocol.python.message import ClientDisconnect, ClientShutdownResponse
from scaler.utility.exceptions import ClientQuitException, ClientShutdownException


class ClientDisconnectManager(DisconnectManager):
    def __init__(self):
        self._connector_internal: Optional[AsyncConnector] = None
        self._connector_external: Optional[AsyncConnector] = None

    def register(self, connector_internal: AsyncConnector, connector_external: AsyncConnector):
        self._connector_internal = connector_internal
        self._connector_external = connector_external

    async def on_client_disconnect(self, disconnect: ClientDisconnect):
        await self._connector_external.send(disconnect)

        if disconnect.disconnect_type == ClientDisconnect.DisconnectType.Disconnect:
            raise ClientQuitException("client disconnecting")

    async def on_client_shutdown_response(self, response: ClientShutdownResponse):
        await self._connector_internal.send(response)

        raise ClientShutdownException("cluster shutting down")



import dataclasses
import unittest

from tests.utility import logging_test_name

from scaler.scheduler.object_usage.object_tracker import ObjectTracker, ObjectUsage
from scaler.utility.logging.utility import setup_logger


@dataclasses.dataclass
class Sample(ObjectUsage):
    key: str
    value: str

    def get_object_key(self) -> str:
        return self.key


def sample_ready(obj: Sample):
    print(f"obj {obj.get_object_key()} is ready")


class TestObjectUsage(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)

    def test_object_usage(self):
        object_usage: ObjectTracker[str, Sample] = ObjectTracker("sample", sample_ready)

        object_usage.add_object(Sample("a", "value1"))
        object_usage.add_object(Sample("b", "value2"))

        object_usage.add_blocks_for_one_object("a", {"1", "2"})
        object_usage.add_blocks_for_one_object("b", {"2", "3", "4"})

        object_usage.remove_blocks({"1"})
        self.assertEqual(object_usage.object_count(), 2)

        object_usage.remove_blocks({"3"})
        self.assertEqual(object_usage.object_count(), 2)

        object_usage.remove_blocks({"2"})
        self.assertEqual(object_usage.object_count(), 1)

        object_usage.remove_blocks({"4"})
        self.assertEqual(object_usage.object_count(), 0)

        object_usage.remove_blocks({"4"})



import unittest

from tests.utility import logging_test_name

from scaler.utility.logging.utility import setup_logger
from scaler.utility.queues.indexed_queue import IndexedQueue


class TestIndexedQueue(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)

    def test_indexed_queue(self):
        queue = IndexedQueue()
        queue.put(1)
        queue.put(2)
        queue.put(3)
        queue.put(4)
        queue.put(5)
        queue.put(6)

        self.assertEqual(len(queue), 6)

        self.assertTrue(1 in queue)
        self.assertTrue(0 not in queue)

        queue.remove(3)
        self.assertEqual(len(queue), 5)
        self.assertTrue(3 not in queue)

        self.assertListEqual(list(queue), [1, 2, 4, 5, 6])

        self.assertEqual(queue.get(), 1)
        self.assertEqual(queue.get(), 2)
        self.assertEqual(len(queue), 3)



import functools
import os
import random
import time
import unittest
from concurrent.futures import CancelledError

from tests.utility import get_available_tcp_port, logging_test_name

from scaler import Client, SchedulerClusterCombo
from scaler.utility.exceptions import MissingObjects, ProcessorDiedError
from scaler.utility.logging.scoped_logger import ScopedLogger
from scaler.utility.logging.utility import setup_logger


def noop(sec: int):
    return sec * 1


def noop_sleep(sec: int):
    time.sleep(sec)
    return sec


def heavy_function(sec: int, payload: bytes):
    return len(payload) * sec


def raise_exception(foo: int):
    if foo == 11:
        raise ValueError("foo cannot be 100")


class TestClient(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)
        self.address = f"tcp://127.0.0.1:{get_available_tcp_port()}"
        self._workers = 3
        self.cluster = SchedulerClusterCombo(address=self.address, n_workers=self._workers, event_loop="builtin")

    def tearDown(self) -> None:
        self.cluster.shutdown()
        pass

    def test_one_submit(self):
        with Client(self.address) as client:
            with ScopedLogger("submitting 1 task"):
                future1 = client.submit(noop, 1)

            self.assertEqual(future1.result(), 1)

    def test_one_map(self):
        with Client(self.address) as client:
            with ScopedLogger("mapping 1 task"):
                result = client.map(noop, [(1,), (2,)])

            self.assertEqual(result, [1, 2])

    def test_submit_and_map(self):
        with Client(self.address) as client:
            tasks = [random.randint(0, 100) for _ in range(100)]
            with ScopedLogger("submitting task use map"):
                map_results = client.map(noop, [(arg,) for arg in tasks])

            self.assertEqual(map_results, tasks)

            with ScopedLogger("submitting task use submit"):
                futures = [client.submit(noop, i) for i in tasks]
                submit_results = [future.result() for future in futures]

            self.assertEqual(submit_results, tasks)

    def test_noop_submit(self):
        with Client(self.address) as client:
            tasks = [random.randint(0, 100) for _ in range(10000)]
            with ScopedLogger(f"submit {len(tasks)} noop tasks"):
                futures = [client.submit(noop, i) for i in tasks]

            with ScopedLogger(f"gather {len(futures)} results"):
                results = [future.result() for future in futures]

            self.assertEqual(results, tasks)

    def test_noop_map(self):
        with Client(self.address) as client:
            tasks = [random.randint(0, 100) for _ in range(10000)]
            with ScopedLogger(f"submit {len(tasks)} noop tasks"):
                results = client.map(noop, [(i,) for i in tasks])

            self.assertEqual(results, tasks)

    def test_noop_cancel(self):
        with Client(self.address) as client:
            tasks = [10, 1, 1] * 10
            with ScopedLogger(f"submit {len(tasks)} noop and cancel tasks"):
                futures = [client.submit(noop_sleep, i) for i in tasks]
                assert isinstance(futures, list)

                for future in futures:
                    future.cancel()

                time.sleep(3)

        time.sleep(1)

    def test_heavy_function(self):
        with Client(self.address) as client:
            size = 500_000_000
            number_of_tasks = 10000
            tasks = [random.randint(0, 100) for _ in range(number_of_tasks)]
            function = functools.partial(heavy_function, payload=b"1" * size)

            with ScopedLogger(f"submit {len(tasks)} heavy function (500mb) for {number_of_tasks} tasks"):
                results = client.map(function, [(i,) for i in tasks])

            expected = [task * size for task in tasks]
            self.assertEqual(results, expected)

    def test_very_large_payload(self):
        def func(data: bytes):
            return data

        with Client(self.address) as client:
            payload = os.urandom(2**29 + 300)  # 512MB + 300B
            future = client.submit(func, payload)

            result = future.result()

            self.assertTrue(payload == result)

    def test_sleep(self):
        with Client(self.address) as client:
            time.sleep(5)

            tasks = [10, 1, 1] * 10
            # tasks = [10] * 10
            with ScopedLogger(f"submit {len(tasks)} sleep and balance tasks"):
                futures = [client.submit(noop_sleep, i) for i in tasks]

            # time.sleep(60)
            # print(f"number of futures: {len(futures)}")
            # print(f"number of states: {Counter([future._state for future in futures])}")
            with ScopedLogger(f"gather {len(futures)} results"):
                results = [future.result() for future in futures]

            self.assertEqual(results, tasks)

    def test_raise_exception(self):
        with Client(self.address) as client:
            tasks = [i for i in range(100)]
            with ScopedLogger(f"submit {len(tasks)} tasks, raise 1 of the tasks"):
                futures = [client.submit(raise_exception, i) for i in tasks]

            with self.assertRaises(ValueError), ScopedLogger(f"gather {len(futures)} results"):
                _ = [future.result() for future in futures]

    def test_function(self):
        def func_args(a: int, b: int, c: int, d: int = 0):
            return a, b, c, d

        def func_args2(a: int, b: int, *, c: int, d: int = 0):
            return a, b, c, d

        with Client(self.address) as client:
            with ScopedLogger("test mix of positional and keyword arguments and with some arguments default value"):
                self.assertEqual(client.submit(func_args, 1, c=4, b=2).result(), (1, 2, 4, 0))

            with ScopedLogger("test all keyword arguments"):
                self.assertEqual(client.submit(func_args, d=5, b=3, c=1, a=4).result(), (4, 3, 1, 5))

            with ScopedLogger("test mix of positional and keyword arguments with override default value"):
                self.assertEqual(client.submit(func_args, 1, c=4, b=2, d=6).result(), (1, 2, 4, 6))

            with ScopedLogger("test partial function"):
                self.assertEqual(client.submit(functools.partial(func_args, 5, 6), 1, 2).result(), (5, 6, 1, 2))

            with ScopedLogger("test insufficient arguments"), self.assertRaises(TypeError):
                client.submit(func_args, 1)

            with ScopedLogger("test not allow keyword only arguments even assigned"), self.assertRaises(TypeError):
                client.submit(func_args2, 1, c=4, b=2, d=6).result()

            with ScopedLogger("test not allow keyword only arguments"), self.assertRaises(TypeError):
                client.submit(func_args2, a=3, b=4).result()

    def test_map(self):
        def func(x, y):
            return x * y

        with Client(self.address) as client:
            result = client.map(func, [(1, 2), (3, 4), (5, 6), (7, 8)])
            self.assertEqual(result, [2, 12, 30, 56])

    def test_more_tasks(self):
        def func(a):
            time.sleep(random.randint(10, 50))
            return a * 2

        with Client(self.address) as client:
            client.map(func, [(i,) for i in range(self._workers * 2)])

    def test_context_manager(self):
        with Client(self.address) as client:
            self.assertEqual(client.submit(noop, 1).result(), 1)

    def test_no_disconnect(self):
        with Client(self.address) as client:
            self.assertEqual(client.submit(noop, 1).result(), 1)

    def test_extra_disconnects(self):
        with Client(self.address) as client:
            self.assertEqual(client.submit(noop, 1).result(), 1)
            client.disconnect()
            client.disconnect()
            client.disconnect()

    def test_processor_died(self):
        def func():
            time.sleep(1)
            os._exit(1)  # noqa
            return 1

        with Client(self.address) as client:
            with self.assertRaises(ProcessorDiedError):
                client.submit(func).result()

    def test_non_hashable_client(self):
        def func(a):
            return a * 2

        with Client(self.address) as client:
            client.submit(func, [1, 2, 3, 4, 5])

    def test_send_object(self):
        def func(a):
            return len(a)

        with Client(self.address) as client:
            ref1 = client.send_object("abcdef")
            self.assertEqual(client.submit(func, ref1).result(), 6)

            ref2 = client.send_object("123456789")
            self.assertEqual(client.map(func, [(ref1,), (ref2,)]), [6, 9])

    def test_send_object2(self):
        def add(a, b):
            return a + b

        with Client(address=self.address) as client:
            ref = client.send_object([1, 2, 3, 4, 5], name="large_object")

            fut = client.submit(add, ref, [6])
            self.assertEqual(fut.result(), [1, 2, 3, 4, 5, 6])

    def test_scheduler_crash(self):
        CLIENT_TIMEOUT_SECONDS = 5

        with Client(address=self.address, timeout_seconds=CLIENT_TIMEOUT_SECONDS) as client:
            future = client.submit(noop, 10)

            self.cluster._scheduler.kill()

            time.sleep(5)

            with self.assertRaises(TimeoutError):
                future.result()

    def test_responsiveness(self):
        MAX_DELAY_SECONDS = 0.25

        # Makes sure the cluster has the time to start up.
        with Client(self.address) as client:
            client.submit(pow, 1, 1).result()

        try:
            connect_start_time = time.time()
            client = Client(self.address)
            self.assertLess(time.time() - connect_start_time, MAX_DELAY_SECONDS)

            submit_start_time = time.time()
            future = client.submit(pow, 2, 3)
            self.assertLess(time.time() - submit_start_time, MAX_DELAY_SECONDS)

            result_start_time = time.time()
            self.assertEqual(future.result(), 8)
            self.assertLess(time.time() - result_start_time, MAX_DELAY_SECONDS)
        finally:
            disconnect_start_time = time.time()
            client.disconnect()
            self.assertLess(time.time() - disconnect_start_time, MAX_DELAY_SECONDS)

    def test_clear(self):
        with Client(self.address) as client:
            arg_reference = client.send_object(0.5)
            future = client.submit(noop_sleep, arg_reference)

            client.clear()

            # clear() cancels all futures
            with self.assertRaises(CancelledError):
                future.result()
            self.assertTrue(future.cancelled())

            # using an old reference should fail
            with self.assertRaises(MissingObjects):
                client.submit(noop_sleep, arg_reference).result()

            # but new tasks should work fine
            self.assertEqual(client.submit(round, 3.14).result(), 3.0)



import datetime

from nicegui import ui

SLIDING_WINDOW_OPTIONS = {
    datetime.timedelta(minutes=5): "5",
    datetime.timedelta(minutes=10): "10",
    datetime.timedelta(minutes=30): "30",
}
MEMORY_STORE_TIME = max(SLIDING_WINDOW_OPTIONS.keys())

MEMORY_USAGE_SCALE_OPTIONS = {"log": "log", "linear": "linear"}


class Settings:
    def __init__(self):
        self.stream_window = datetime.timedelta(minutes=5)
        self.memory_store_time = MEMORY_STORE_TIME
        self.memory_usage_scale = "log"

    def draw_section(self):
        with ui.card().classes("w-fit").classes("q-mx-auto"):
            ui.label("Sliding Window Length").classes("q-mx-auto")
            ui.toggle(SLIDING_WINDOW_OPTIONS).bind_value(self, "stream_window")

        with ui.card().classes("w-fit").classes("q-mx-auto"):
            ui.label("Memory Store Time").classes("q-mx-auto")
            ui.toggle(SLIDING_WINDOW_OPTIONS).bind_value(self, "memory_store_time")

        with ui.card().classes("w-fit").classes("q-mx-auto"):
            ui.label("Memory Usage Scale").classes("q-mx-auto")
            ui.toggle(MEMORY_USAGE_SCALE_OPTIONS).bind_value(self, "memory_usage_scale")

    @staticmethod
    def max_window_size() -> datetime.timedelta:
        return max(SLIDING_WINDOW_OPTIONS.keys())



import asyncio
import logging
import multiprocessing
import signal
from typing import Optional, Tuple

import zmq.asyncio

from scaler.io.async_connector import AsyncConnector
from scaler.io.config import PROFILING_INTERVAL_SECONDS
from scaler.protocol.python.message import (
    ClientDisconnect,
    DisconnectRequest,
    ObjectInstruction,
    ObjectResponse,
    Task,
    TaskCancel,
    WorkerHeartbeatEcho,
)
from scaler.protocol.python.mixins import Message
from scaler.utility.event_loop import create_async_loop_routine, register_event_loop
from scaler.utility.exceptions import ClientShutdownException
from scaler.utility.logging.utility import setup_logger
from scaler.utility.zmq_config import ZMQConfig
from scaler.worker.agent.heartbeat_manager import VanillaHeartbeatManager
from scaler.worker.agent.object_tracker import VanillaObjectTracker
from scaler.worker.agent.processor_manager import VanillaProcessorManager
from scaler.worker.agent.profiling_manager import VanillaProfilingManager
from scaler.worker.agent.task_manager import VanillaTaskManager
from scaler.worker.agent.timeout_manager import VanillaTimeoutManager


class Worker(multiprocessing.get_context("spawn").Process):  # type: ignore
    def __init__(
        self,
        event_loop: str,
        name: str,
        address: ZMQConfig,
        io_threads: int,
        heartbeat_interval_seconds: int,
        garbage_collect_interval_seconds: int,
        trim_memory_threshold_bytes: int,
        task_timeout_seconds: int,
        death_timeout_seconds: int,
        hard_processor_suspend: bool,
        logging_paths: Tuple[str, ...],
        logging_level: str,
    ):
        multiprocessing.Process.__init__(self, name="Agent")

        self._event_loop = event_loop
        self._name = name
        self._address = address
        self._io_threads = io_threads

        self._heartbeat_interval_seconds = heartbeat_interval_seconds
        self._garbage_collect_interval_seconds = garbage_collect_interval_seconds
        self._trim_memory_threshold_bytes = trim_memory_threshold_bytes
        self._task_timeout_seconds = task_timeout_seconds
        self._death_timeout_seconds = death_timeout_seconds
        self._hard_processor_suspend = hard_processor_suspend

        self._logging_paths = logging_paths
        self._logging_level = logging_level

        self._context: Optional[zmq.asyncio.Context] = None
        self._connector_external: Optional[AsyncConnector] = None
        self._task_manager: Optional[VanillaTaskManager] = None
        self._heartbeat_manager: Optional[VanillaHeartbeatManager] = None
        self._profiling_manager: Optional[VanillaProfilingManager] = None
        self._processor_manager: Optional[VanillaProcessorManager] = None

    @property
    def identity(self):
        return self._connector_external.identity

    def run(self) -> None:
        self.__initialize()
        self.__run_forever()

    def __initialize(self):
        setup_logger()
        register_event_loop(self._event_loop)

        self._context = zmq.asyncio.Context()
        self._connector_external = AsyncConnector(
            context=self._context,
            name=self.name,
            socket_type=zmq.DEALER,
            address=self._address,
            bind_or_connect="connect",
            callback=self.__on_receive_external,
            identity=None,
        )

        self._heartbeat_manager = VanillaHeartbeatManager()
        self._profiling_manager = VanillaProfilingManager()
        self._task_manager = VanillaTaskManager(task_timeout_seconds=self._task_timeout_seconds)
        self._timeout_manager = VanillaTimeoutManager(death_timeout_seconds=self._death_timeout_seconds)
        self._processor_manager = VanillaProcessorManager(
            context=self._context,
            event_loop=self._event_loop,
            garbage_collect_interval_seconds=self._garbage_collect_interval_seconds,
            trim_memory_threshold_bytes=self._trim_memory_threshold_bytes,
            hard_processor_suspend=self._hard_processor_suspend,
            logging_paths=self._logging_paths,
            logging_level=self._logging_level,
        )
        self._object_tracker = VanillaObjectTracker()

        # register
        self._task_manager.register(connector=self._connector_external, processor_manager=self._processor_manager)
        self._heartbeat_manager.register(
            connector_external=self._connector_external,
            worker_task_manager=self._task_manager,
            timeout_manager=self._timeout_manager,
            processor_manager=self._processor_manager,
        )
        self._processor_manager.register(
            heartbeat=self._heartbeat_manager,
            task_manager=self._task_manager,
            profiling_manager=self._profiling_manager,
            object_tracker=self._object_tracker,
            connector_external=self._connector_external,
        )

        self._loop = asyncio.get_event_loop()
        self.__register_signal()
        self._task = self._loop.create_task(self.__get_loops())

    async def __on_receive_external(self, message: Message):
        if isinstance(message, WorkerHeartbeatEcho):
            await self._heartbeat_manager.on_heartbeat_echo(message)
            return

        if isinstance(message, Task):
            await self._task_manager.on_task_new(message)
            return

        if isinstance(message, TaskCancel):
            await self._task_manager.on_cancel_task(message)
            return

        if isinstance(message, ObjectInstruction):
            await self._processor_manager.on_object_instruction(message)
            return

        if isinstance(message, ObjectResponse):
            await self._processor_manager.on_object_response(message)
            return

        if isinstance(message, ClientDisconnect):
            if message.disconnect_type == ClientDisconnect.DisconnectType.Shutdown:
                raise ClientShutdownException("received client shutdown, quitting")
            logging.error(f"Worker received invalid ClientDisconnect type, ignoring {message=}")
            return

        raise TypeError(f"Unknown {message=}")

    async def __get_loops(self):
        await self._processor_manager.initialize()

        try:
            await asyncio.gather(
                create_async_loop_routine(self._connector_external.routine, 0),
                create_async_loop_routine(self._heartbeat_manager.routine, self._heartbeat_interval_seconds),
                create_async_loop_routine(self._timeout_manager.routine, 1),
                create_async_loop_routine(self._task_manager.routine, 0),
                create_async_loop_routine(self._profiling_manager.routine, PROFILING_INTERVAL_SECONDS),
                create_async_loop_routine(self._processor_manager.routine, 0),
            )
        except asyncio.CancelledError:
            pass
        except (ClientShutdownException, TimeoutError) as e:
            logging.info(f"Worker[{self.pid}]: {str(e)}")
        except Exception as e:
            logging.exception(f"Worker[{self.pid}]: failed with unhandled exception:\n{(e)}")

        await self._connector_external.send(DisconnectRequest.new_msg(self._connector_external.identity))

        self._connector_external.destroy()
        self._processor_manager.destroy("quited")
        logging.info(f"Worker[{self.pid}]: quited")

    def __run_forever(self):
        self._loop.run_until_complete(self._task)

    def __register_signal(self):
        self._loop.add_signal_handler(signal.SIGINT, self.__destroy)

    def __destroy(self):
        self._task.cancel()



import functools
import inspect
import logging
import typing

from scaler.utility.logging.scoped_logger import ScopedLogger


def log_function(level_number: int = 2, logging_level: int = logging.INFO) -> typing.Callable:
    def decorator(func: typing.Callable) -> typing.Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with ScopedLogger(
                f"execute {func.__name__} at {get_caller_location(level_number)}", logging_level=logging_level
            ):
                return func(*args, **kwargs)

        return wrapper

    return decorator


def get_caller_location(stack_level: int):
    caller = inspect.getframeinfo(inspect.stack()[stack_level][0])
    return f"{caller.filename}:{caller.lineno}"



import functools
import pdb
import sys
from typing import Callable


def pdb_wrapped(func: Callable):
    @functools.wraps(func)
    def pdb_wrapper(*args, **kwargs):
        try:
            exit_code = func(*args, **kwargs)
            sys.exit(exit_code)

        except Exception:
            ex_type, value, tb = sys.exc_info()
            pdb.post_mortem(tb)
            raise

    return pdb_wrapper



import dataclasses
import enum
from typing import List, Tuple

from scaler.protocol.capnp._python import _common  # noqa
from scaler.protocol.python.mixins import Message


class TaskStatus(enum.Enum):
    # task is accepted by scheduler, but will have below status
    Success = _common.TaskStatus.success  # if submit and task is done and get result
    Failed = _common.TaskStatus.failed  # if submit and task is failed on worker
    Canceled = _common.TaskStatus.canceled  # if submit and task is canceled
    NotFound = _common.TaskStatus.notFound  # if submit and task is not found in scheduler
    WorkerDied = (
        _common.TaskStatus.workerDied
    )  # if submit and worker died (only happened when scheduler keep_task=False)
    NoWorker = _common.TaskStatus.noWorker  # if submit and scheduler is full (not implemented yet)

    # below are only used for monitoring channel, not sent to client
    Inactive = _common.TaskStatus.inactive  # task is scheduled but not allocate to worker
    Running = _common.TaskStatus.running  # task is running in worker
    Canceling = _common.TaskStatus.canceling  # task is canceling (can be in Inactive or Running state)


@dataclasses.dataclass
class ObjectContent(Message):
    class ObjectContentType(enum.Enum):
        # FIXME: Pycapnp does not support assignment of raw enum values when the enum is itself declared within a list.
        # However, assigning the enum's string value works.
        # See https://github.com/capnproto/pycapnp/issues/374

        Serializer = "serializer"
        Object = "object"

    def __init__(self, msg):
        super().__init__(msg)

    @property
    def object_ids(self) -> Tuple[bytes, ...]:
        return tuple(self._msg.objectIds)

    @property
    def object_types(self) -> Tuple[ObjectContentType, ...]:
        return tuple(ObjectContent.ObjectContentType(object_type._as_str()) for object_type in self._msg.objectTypes)

    @property
    def object_names(self) -> Tuple[bytes, ...]:
        return tuple(self._msg.objectNames)

    @property
    def object_bytes(self) -> Tuple[List[bytes], ...]:
        return tuple(self._msg.objectBytes)

    @staticmethod
    def new_msg(
        object_ids: Tuple[bytes, ...],
        object_types: Tuple[ObjectContentType, ...] = tuple(),
        object_names: Tuple[bytes, ...] = tuple(),
        object_bytes: Tuple[List[bytes], ...] = tuple(),
    ) -> "ObjectContent":
        return ObjectContent(
            _common.ObjectContent(
                objectIds=list(object_ids),
                objectTypes=[object_type.value for object_type in object_types],
                objectNames=list(object_names),
                objectBytes=tuple(object_bytes),
            )
        )

    def get_message(self):
        return self._msg



import pickle
from typing import Any

import cloudpickle

from scaler.client.serializer.mixins import Serializer


class DefaultSerializer(Serializer):
    @staticmethod
    def serialize(obj: Any) -> bytes:
        return cloudpickle.dumps(obj, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def deserialize(payload: bytes) -> Any:
        return cloudpickle.loads(payload)



import datetime
from typing import Any, Dict, Optional

from nicegui import ui

from scaler.protocol.python.message import StateTask
from scaler.ui.setting_page import Settings
from scaler.ui.utility import format_timediff, get_bounds, make_tick_text, make_ticks
from scaler.utility.formatter import format_bytes
from scaler.utility.metadata.profile_result import ProfileResult

CHART_NAME = "Memory Usage"
X_AXIS_GRID_LINES = False

Y_AXIS_TICK_VALS = [0, 1024, 1024 * 1024, 1024 * 1024 * 1024, 1024 * 1024 * 1024 * 1024]
Y_AXIS_TICK_TEXT = ["0", "1KB", "1MB", "1GB", "1TB"]
Y_AXIS_GRID_LINES = True


class MemoryChart:
    def __init__(self):
        self._figure = {}
        self._plot = None
        self._plot_data: Dict[str, Any] = {}

        self._settings: Optional[Settings] = None

        self._start_time = datetime.datetime.now() - datetime.timedelta(minutes=30)

    def setup_memory_chart(self, settings: Settings):
        with ui.card().classes("w-full").style("height: 30vh"):
            self._plot_data = {
                "type": "scatter",
                "fill": "tozeroy",
                "fillcolor": "rgba(0,0,255,1)",
                "mode": "none",
                "name": "",
                "x": [],
                "y": [],
                "hovertemplate": [],
            }
            fig = {
                "data": [self._plot_data],
                "layout": {
                    "title": {"text": CHART_NAME},
                    "autosize": True,
                    "margin": {"l": 163},
                    "xaxis": {
                        "autorange": False,
                        "range": [0, 300],
                        "showgrid": X_AXIS_GRID_LINES,
                        "tickmode": "array",
                        "tickvals": [0, 50, 100, 150, 200, 250, 300],
                        "ticktext": [-300, -250, -200, -150, -100, -50, 0],
                        "zeroline": False,
                    },
                    "yaxis": {
                        "tickvals": Y_AXIS_TICK_VALS,
                        "ticktext": Y_AXIS_TICK_TEXT,
                        "autorange": True,
                        "automargin": True,
                        "rangemode": "nonnegative",
                        "showgrid": Y_AXIS_GRID_LINES,
                        "type": "log",
                    },
                },
            }
            self._figure = fig
            self._plot = ui.plotly(self._figure).classes("w-full h-full")
            self._settings = settings

    def handle_task_state(self, state: StateTask):
        """
        Only completed tasks have profiling data.
        Use this data to fill in history.
        """

        if state.metadata == b"":
            return

        profile_result = ProfileResult.deserialize(state.metadata)

        worker_memory = profile_result.memory_peak
        worker_duration = profile_result.duration_s

        if worker_memory == 0:
            return

        self.__add_memory_usage(worker_duration, worker_memory)

    def update_plot(self):
        now = datetime.datetime.now()
        self.__render_plot(now)

    def __add_memory_usage(self, time_taken: float, memory_usage: int):
        now = datetime.datetime.now()
        current_time = format_timediff(self._start_time, now)
        task_start_time = current_time - time_taken

        # find index we need to start changing memory use from
        insert_index = len(self._plot_data["x"])
        while insert_index > 0 and self._plot_data["x"][insert_index - 1] > task_start_time:
            insert_index -= 1

        # insert points to mark change
        self._plot_data["x"].insert(insert_index, task_start_time)
        self._plot_data["x"].insert(insert_index, task_start_time - 0.01)

        prev_mem = 0
        if insert_index < len(self._plot_data["y"]):
            prev_mem = self._plot_data["y"][insert_index]

        self._plot_data["y"].insert(insert_index, prev_mem + memory_usage)
        self._plot_data["y"].insert(insert_index, prev_mem)

        self._plot_data["hovertemplate"].insert(insert_index, format_bytes(prev_mem + memory_usage))
        self._plot_data["hovertemplate"].insert(insert_index, format_bytes(prev_mem))

        # fill future overlapping points with the additional memory use
        i = insert_index + 2
        while i < len(self._plot_data["x"]):
            self._plot_data["y"][i] += memory_usage
            self._plot_data["hovertemplate"][i] = format_bytes(self._plot_data["y"][i])
            i += 1

        # mark end of this task's memory use. always going to be the latest task we have so far.
        self._plot_data["y"].append(memory_usage)
        self._plot_data["hovertemplate"].append(format_bytes(memory_usage))
        self._plot_data["x"].append(current_time - 0.01)

        self._plot_data["y"].append(0)
        self._plot_data["hovertemplate"].append(format_bytes(0))
        self._plot_data["x"].append(current_time)

    def __render_plot(self, now: datetime.datetime):
        lower_bound, upper_bound = get_bounds(now, self._start_time, self._settings)

        ticks = make_ticks(lower_bound, upper_bound)
        tick_text = make_tick_text(int(self._settings.stream_window.total_seconds()))

        self._figure["layout"]["xaxis"]["range"] = [lower_bound, upper_bound]
        self._figure["layout"]["xaxis"]["tickvals"] = ticks
        self._figure["layout"]["xaxis"]["ticktext"] = tick_text
        self._figure["layout"]["yaxis"]["type"] = self._settings.memory_usage_scale

        self._plot.update()



import os

# ==============
# SYSTEM OPTIONS

# object clean up time interval
CLEANUP_INTERVAL_SECONDS = 1

# status report interval, used by poke or scaled monitor
STATUS_REPORT_INTERVAL_SECONDS = 1

# number of seconds for profiling
PROFILING_INTERVAL_SECONDS = 1

# cap'n proto only allow Data/Text/Blob size to be as big as 500MB
CAPNP_DATA_SIZE_LIMIT = 2**29 - 1

# message size limitation, max can be 2**64
CAPNP_MESSAGE_SIZE_LIMIT = 2**64 - 1

# ==========================
# SCHEDULER SPECIFIC OPTIONS

# number of threads for zmq socket to handle
DEFAULT_IO_THREADS = 1

# if all workers are full and busy working, this option determine how many additional tasks scheduler can receive and
# queued, if additional number of tasks received exceeded this number, scheduler will reject tasks
DEFAULT_MAX_NUMBER_OF_TASKS_WAITING = -1

# if didn't receive heartbeat for following seconds, then scheduler will treat worker as dead and reschedule unfinished
# tasks for this worker
DEFAULT_WORKER_TIMEOUT_SECONDS = 60

# if didn't receive heartbeat for following seconds, then scheduler will treat client as dead and cancel remaining
# tasks for this client
DEFAULT_CLIENT_TIMEOUT_SECONDS = 60

# number of seconds for load balance, if value is 0 means disable load balance
# FIXME: load balancing is currently disabled by default as it's causing some issues under heavy load.
DEFAULT_LOAD_BALANCE_SECONDS = 0

# when load balance advice happened repeatedly and always be the same, we issue load balance request when exact repeated
# times happened
DEFAULT_LOAD_BALANCE_TRIGGER_TIMES = 2

# number of tasks can be queued to each worker on scheduler side
DEFAULT_PER_WORKER_QUEUE_SIZE = 1000

# =======================
# WORKER SPECIFIC OPTIONS

# number of workers, echo worker use 1 process
DEFAULT_NUMBER_OF_WORKER = os.cpu_count() - 1

# number of seconds that worker agent send heartbeat to scheduler
DEFAULT_HEARTBEAT_INTERVAL_SECONDS = 2

# number of seconds the object cache kept in worker's memory
DEFAULT_OBJECT_RETENTION_SECONDS = 60

# number of seconds worker doing garbage collection
DEFAULT_GARBAGE_COLLECT_INTERVAL_SECONDS = 30

# number of bytes threshold for worker process that trigger deep garbage collection
DEFAULT_TRIM_MEMORY_THRESHOLD_BYTES = 1024 * 1024 * 1024

# default task timeout seconds, 0 means never timeout
DEFAULT_TASK_TIMEOUT_SECONDS = 0

# number of seconds that worker agent wait for processor to finish before killing it
DEFAULT_PROCESSOR_KILL_DELAY_SECONDS = 3

# number of seconds without scheduler contact before worker shuts down
DEFAULT_WORKER_DEATH_TIMEOUT = 5 * 60

# the global client name for get log, right now, all client shared remote log, task need have client information to
# deliver log
DUMMY_CLIENT = b"dummy_client"

# if true, suspended worker's processors will be actively suspended with a SIGTSTP signal, otherwise a synchronization
# event will be used.
DEFAULT_HARD_PROCESSOR_SUSPEND = False



import dataclasses
import enum
from typing import Optional


class ZMQType(enum.Enum):
    inproc = "inproc"
    ipc = "ipc"
    tcp = "tcp"

    @staticmethod
    def allowed_types():
        return {t.value for t in ZMQType}


@dataclasses.dataclass
class ZMQConfig:
    type: ZMQType
    host: str
    port: Optional[int] = None

    def __post_init__(self):
        if not isinstance(self.type, ZMQType):
            raise TypeError(f"Invalid zmq type {self.type}, available types are: {ZMQType.allowed_types()}")

        if not isinstance(self.host, str):
            raise TypeError(f"Host should be string, given {self.host}")

        if self.port is None:
            if self.type == ZMQType.tcp:
                raise ValueError(f"type {self.type.value} should have `port`")
        else:
            if self.type in {ZMQType.inproc, ZMQType.ipc}:
                raise ValueError(f"type {self.type.value} should not have `port`")

            if not isinstance(self.port, int):
                raise TypeError(f"Port should be integer, given {self.port}")

    def to_address(self):
        if self.type == ZMQType.tcp:
            return f"tcp://{self.host}:{self.port}"

        if self.type in {ZMQType.inproc, ZMQType.ipc}:
            return f"{self.type.value}://{self.host}"

        raise TypeError(f"Unsupported ZMQ type: {self.type}")

    @staticmethod
    def from_string(string: str) -> "ZMQConfig":
        if "://" not in string:
            raise ValueError("valid ZMQ config should be like tcp://127.0.0.1:12345")

        socket_type, host_port = string.split("://", 1)
        if socket_type not in ZMQType.allowed_types():
            raise ValueError(f"supported ZMQ types are: {ZMQType.allowed_types()}")

        socket_type_enum = ZMQType(socket_type)
        if socket_type_enum in {ZMQType.inproc, ZMQType.ipc}:
            host = host_port
            port_int = None
        elif socket_type_enum == ZMQType.tcp:
            host, port = host_port.split(":")
            try:
                port_int = int(port)
            except ValueError:
                raise ValueError(f"cannot convert '{port}' to port number")
        else:
            raise ValueError(f"Unsupported ZMQ type: {socket_type}")

        return ZMQConfig(socket_type_enum, host, port_int)



import asyncio
import unittest

from tests.utility import logging_test_name

from scaler.utility.logging.utility import setup_logger
from scaler.utility.queues.async_indexed_queue import AsyncIndexedQueue


class TestAsyncIndexedQueue(unittest.TestCase):
    def setUp(self) -> None:
        setup_logger()
        logging_test_name(self)

    def test_async_indexed_queue(self):
        async def async_test():
            queue = AsyncIndexedQueue()
            await queue.put(1)
            await queue.put(2)
            await queue.put(3)
            await queue.put(4)
            await queue.put(5)
            await queue.put(6)

            queue.remove(1)
            queue.remove(3)
            queue.remove(6)
            self.assertEqual(queue.qsize(), 3)

            self.assertEqual(await queue.get(), 2)
            self.assertEqual(await queue.get(), 4)
            self.assertEqual(await queue.get(), 5)
            self.assertEqual(queue.qsize(), 0)
            self.assertTrue(not queue)
            self.assertTrue(queue.empty())

        asyncio.run(async_test())

    def test_duplicated_items(self):
        async def async_test():
            queue = AsyncIndexedQueue(3)
            await queue.put(1)
            await queue.put(1)

        with self.assertRaises(KeyError):
            asyncio.run(async_test())



import datetime
import logging
import time
from typing import Optional


class ScopedLogger:
    def __init__(self, message: str, logging_level=logging.INFO):
        self.timer = TimedLogger(message=message, logging_level=logging_level)

    def __enter__(self):
        self.timer.begin()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.timer.end()


class TimedLogger:
    def __init__(self, message: str, logging_level=logging.INFO):
        self.message = message
        self.logging_level = logging_level
        self.timer: Optional[int] = None

    def begin(self):
        self.timer = time.perf_counter_ns()
        logging.log(self.logging_level, f"beginning {self.message}")

    def end(self):
        elapsed = time.perf_counter_ns() - self.timer
        offset = datetime.timedelta(
            seconds=int(elapsed / 1e9), milliseconds=int(elapsed % 1e9 / 1e6), microseconds=int(elapsed % 1e6 / 1e3)
        )
        logging.log(self.logging_level, f"completed {self.message} in {offset}")



import datetime
from typing import List, Tuple

from scaler.ui.setting_page import Settings


def format_timediff(a: datetime.datetime, b: datetime.datetime) -> float:
    return (b - a).total_seconds()


def format_worker_name(worker_name: str) -> str:
    pid, host, hash_code = worker_name.split("|")
    return f"{host}|{pid}"


def get_bounds(now: datetime.datetime, start_time: datetime.datetime, settings: Settings) -> Tuple[int, int]:
    upper_range = now - start_time
    lower_range = upper_range - settings.stream_window

    bound_upper_seconds = max(upper_range.seconds, settings.stream_window.seconds)
    bound_lower_seconds = 0 if bound_upper_seconds == settings.stream_window.seconds else lower_range.seconds

    return bound_lower_seconds, bound_upper_seconds


def make_ticks(lower_bound: int, upper_bound: int) -> List[int]:
    distance = (upper_bound - lower_bound) // 6
    return list(range(lower_bound, upper_bound + 1, distance))


def make_tick_text(window_length: int) -> List[int]:
    upper = 0
    lower = -1 * window_length
    distance = (upper - lower) // 6
    return list(range(lower, upper + 1, distance))



import logging
import socket
import unittest


def get_available_tcp_port(hostname: str = "127.0.0.1") -> int:
    with socket.socket(socket.AddressFamily.AF_INET, socket.SocketKind.SOCK_STREAM) as sock:
        sock.bind((hostname, 0))
        return sock.getsockname()[1]


def logging_test_name(obj: unittest.TestCase):
    logging.info(f"{obj.__class__.__name__}:{obj._testMethodName} ==============================================")



import dataclasses
import struct

from scaler.protocol.python.message import Task


@dataclasses.dataclass
class TaskFlags:
    profiling: bool = dataclasses.field(default=True)
    priority: int = dataclasses.field(default=0)

    FORMAT = "!?i"

    def serialize(self) -> bytes:
        return struct.pack(TaskFlags.FORMAT, self.profiling, self.priority)

    @staticmethod
    def deserialize(data: bytes) -> "TaskFlags":
        return TaskFlags(*struct.unpack(TaskFlags.FORMAT, data))


def retrieve_task_flags_from_task(task: Task) -> TaskFlags:
    if task.metadata == b"":
        return TaskFlags()

    try:
        return TaskFlags.deserialize(task.metadata)
    except struct.error:
        raise ValueError(f"unexpected metadata value (expected {TaskFlags.__name__}).")



import collections
from typing import Callable


class EventList(collections.UserList):
    """A list that emits events when it is modified."""

    def __init__(self, initlist=None):
        super().__init__(initlist=initlist)
        self._callbacks = []

    def add_update_callback(self, callback: Callable[["EventList"], None]):
        self._callbacks.append(callback)

    def __setitem__(self, i, item):
        super().__setitem__(i, item)
        self._list_updated()

    def __delitem__(self, i):
        super().__delitem__(i)
        self._list_updated()

    def __add__(self, other):
        super().__add__(other)
        self._list_updated()

    def __iadd__(self, other):
        super().__iadd__(other)
        self._list_updated()
        return self

    def append(self, item):
        super().append(item)
        self._list_updated()

    def insert(self, i, item):
        super().insert(i, item)
        self._list_updated()

    def pop(self, i: int = -1):
        v = super().pop(i)
        self._list_updated()
        return v

    def remove(self, item):
        super().remove(item)
        self._list_updated()

    def clear(self) -> None:
        super().clear()
        self._list_updated()

    def sort(self, /, *args, **kwargs):
        super().sort(*args, **kwargs)
        self._list_updated()

    def extend(self, other) -> None:
        super().extend(other)
        self._list_updated()

    def _list_updated(self):
        for callback in self._callbacks:
            callback(self)



import math
from typing import Dict, List, Optional, Set

from scaler.scheduler.allocators.mixins import TaskAllocator
from scaler.utility.queues.async_priority_queue import AsyncPriorityQueue
from scaler.utility.queues.indexed_queue import IndexedQueue


class QueuedAllocator(TaskAllocator):
    def __init__(self, max_tasks_per_worker: int):
        self._max_tasks_per_worker = max_tasks_per_worker
        self._workers_to_task_ids: Dict[bytes, IndexedQueue] = dict()
        self._task_id_to_worker: Dict[bytes, bytes] = {}

        self._worker_queue: AsyncPriorityQueue = AsyncPriorityQueue()

    async def add_worker(self, worker: bytes) -> bool:
        if worker in self._workers_to_task_ids:
            return False

        self._workers_to_task_ids[worker] = IndexedQueue()
        await self._worker_queue.put([0, worker])
        return True

    def remove_worker(self, worker: bytes) -> List[bytes]:
        if worker not in self._workers_to_task_ids:
            return []

        self._worker_queue.remove(worker)

        task_ids = list(self._workers_to_task_ids.pop(worker))
        for task_id in task_ids:
            self._task_id_to_worker.pop(task_id)
        return task_ids

    def get_worker_ids(self) -> Set[bytes]:
        return set(self._workers_to_task_ids.keys())

    def get_worker_by_task_id(self, task_id: bytes) -> bytes:
        return self._task_id_to_worker.get(task_id, b"")

    def balance(self) -> Dict[bytes, List[bytes]]:
        """Returns, for every worker, the list of tasks to balance out."""

        balance_count = self.__get_balance_count_by_worker()

        balance_result = {}

        for worker, count in balance_count.items():
            if count == 0:
                continue

            tasks = list(self._workers_to_task_ids[worker])
            balance_result[worker] = tasks[-count:]  # balance out the most recently queued tasks

        return balance_result

    def __get_balance_count_by_worker(self) -> Dict[bytes, int]:
        """Returns, for every worker, the number of tasks to balance out."""

        queued_tasks_per_worker = {
            worker: max(0, len(tasks) - 1) for worker, tasks in self._workers_to_task_ids.items()
        }

        any_worker_has_queued_task = any(queued_tasks_per_worker.values())

        if not any_worker_has_queued_task:
            return {}

        number_of_idle_workers = sum(1 for tasks in self._workers_to_task_ids.values() if len(tasks) == 0)

        if number_of_idle_workers == 0:
            return {}

        mean_queued = math.ceil(sum(queued_tasks_per_worker.values()) / len(queued_tasks_per_worker))

        balance_count = {worker: max(0, count - mean_queued) for worker, count in queued_tasks_per_worker.items()}

        over_mean_advice_total = sum(balance_count.values())
        minimal_allocate = min(number_of_idle_workers, sum(queued_tasks_per_worker.values()))

        if over_mean_advice_total >= minimal_allocate:
            return balance_count

        total_to_be_balance = minimal_allocate - over_mean_advice_total
        for worker, count in queued_tasks_per_worker.items():
            assert total_to_be_balance >= 0, "total_to_be_balance must be positive"
            if total_to_be_balance == 0:
                break

            leftover = count - balance_count[worker]
            if leftover < 1:
                continue

            to_to_balance = min(leftover, total_to_be_balance)
            balance_count[worker] += to_to_balance
            total_to_be_balance -= to_to_balance

        return balance_count

    async def assign_task(self, task_id: bytes) -> Optional[bytes]:
        if task_id in self._task_id_to_worker:
            return self._task_id_to_worker[task_id]

        count, worker = await self._worker_queue.get()
        if count == self._max_tasks_per_worker:
            await self._worker_queue.put([count, worker])
            return None

        self._workers_to_task_ids[worker].put(task_id)
        self._task_id_to_worker[task_id] = worker
        await self._worker_queue.put([count + 1, worker])
        return worker

    def remove_task(self, task_id: bytes) -> Optional[bytes]:
        if task_id not in self._task_id_to_worker:
            return None

        worker = self._task_id_to_worker.pop(task_id)
        self._workers_to_task_ids[worker].remove(task_id)

        self._worker_queue.decrease_priority(worker)
        return worker

    def get_assigned_worker(self, task_id: bytes) -> Optional[bytes]:
        if task_id not in self._task_id_to_worker:
            return None

        return self._task_id_to_worker[task_id]

    def has_available_worker(self) -> bool:
        if not len(self._worker_queue):
            return False

        count = self._worker_queue.max_priority()
        if count == self._max_tasks_per_worker:
            return False

        return True

    def statistics(self) -> Dict:
        return {
            worker: {"free": self._max_tasks_per_worker - len(tasks), "sent": len(tasks)}
            for worker, tasks in self._workers_to_task_ids.items()
        }
